/* tile definition file. This file is linked into the "tiletrans" utility,
 * which separates the tiles and generates the vultures_gametiles.{c,h} code
 *
 * Copyright (c) 2005 Daniel Thaler
 * This file may be freely distributed under the Nethack license*/

#include "tiletrans.h"
#include "config.h"
#include "monsym.h"
#include "objclass.h"


#define MON_TINY_DX 57
#define MON_TINY_DY 62
#define MON_DX 116
#define MON_DY 126
#define MON_LARGE_DX 154
#define MON_LARGE_DY 165
#define MON_HUGE_DX 232
#define MON_HUGE_DY 252
#define FIG_DX 58
#define FIG_DY 63
#define FIG_LARGE_DX 77
#define FIG_LARGE_DY 84
#define FIG_HUGE_DX 116
#define FIG_HUGE_DY 126
#define CMAP_DX 116
#define CMAP_DY 126
#define OBJ_DX 96
#define OBJ_DY 102


const struct floorstyles fstyles[] = {
/* basename           x  y */
{"FLOOR_COBBLESTONE", 3, 3},
{"FLOOR_ROUGH",       3, 3},
{"FLOOR_CERAMIC",     3, 3},
{"FLOOR_LAVA",        3, 3},
{"FLOOR_WATER",       3, 3},
{"FLOOR_ICE",         3, 3},
{"FLOOR_MURAL",       3, 2},
{"FLOOR_MURAL2",      3, 2},
{"FLOOR_CARPET",      3, 2},
{"FLOOR_MOSS_COVERED",3, 3},
{"FLOOR_MARBLE",      3, 3},
{"FLOOR_ROUGH_LIT",   3, 3},
{"FLOOR_AIR",         3, 3},
{"FLOOR_DARK",        1, 1}
};


const char * floor_edges[] = {
	"FLOOR_EDGE_COBBLESTONE"
};



const char * wall_styles[] = {
	"WALL_BRICK",
	"WALL_BRICK_BANNER",
	"WALL_BRICK_PAINTING",
	"WALL_BRICK_POCKET",
	"WALL_BRICK_PILLAR",
	"WALL_MARBLE",
	"WALL_VINE_COVERED",
	"WALL_STUCCO",
	"WALL_ROUGH",
	"WALL_DARK",
	"WALL_LIGHT"
};


/* classdefmaps assign a default tile to an entire class of monsters or objects */
const struct classdefmap objclassmap[] = {
{ILLOBJ_CLASS,  "STRANGE_OBJECT"},
{WEAPON_CLASS,  "SPEAR"},
{ARMOR_CLASS,   "HELMET"},
{RING_CLASS,    "RIN_ADORNMENT"},
{AMULET_CLASS,  "AMULET_OF_ESP"},
{TOOL_CLASS,    "STRANGE_OBJECT"},
{FOOD_CLASS,    "FOOD_RATION"},
{POTION_CLASS,  "POT_WATER"},
{SCROLL_CLASS,  "SCR_ENCHANT_ARMOR"},
{SPBOOK_CLASS,  "SPE_DIG"},
{WAND_CLASS,    "WAN_LIGHT"},
{COIN_CLASS,    "GOLD_PIECE"},
{GEM_CLASS,     "DIAMOND"},
{ROCK_CLASS,    "BOULDER"},
{BALL_CLASS,    "STRANGE_OBJECT"},
{CHAIN_CLASS,   "STRANGE_OBJECT"},
{VENOM_CLASS,   "POT_WATER"}
};

const struct classdefmap monclassmap[] = {
{S_ANT,           "PM_SOLDIER_ANT"},
{S_BLOB,          "PM_GREEN_SLIME"},
{S_COCKATRICE,    "PM_FLOATING_EYE"},
{S_DOG,           "PM_WOLF"},
{S_EYE,           "PM_FLOATING_EYE"},
{S_FELINE,        "PM_LARGE_CAT"},
{S_GREMLIN,       "PM_GNOME"},
{S_HUMANOID,      "PM_GOBLIN"},
{S_IMP,           "PM_GOBLIN"},
{S_JELLY,         "PM_GREEN_SLIME"},
{S_KOBOLD,        "PM_GOBLIN"},
{S_LEPRECHAUN,    "PM_GNOME"},
{S_MIMIC,         "PM_LARGE_MIMIC"},
{S_NYMPH,         "PM_WOOD_NYMPH"},
{S_ORC,           "PM_ORC"},
{S_QUADRUPED,     "PM_HORSE"},
{S_RODENT,        "PM_GIANT_RAT"},
{S_SPIDER,        "PM_GIANT_SPIDER"},
{S_UNICORN,       "PM_HORSE"},
{S_ANGEL,         "PM_ANGEL"},
{S_BAT,           "PM_VAMPIRE_BAT"},
{S_CENTAUR,       "PM_FOREST_CENTAUR"},
{S_DRAGON,        "PM_BLACK_DRAGON"},
{S_FUNGUS,        "PM_LICHEN"},
{S_GNOME,         "PM_GNOME"},
{S_GIANT,         "PM_GIANT"},
{S_JABBERWOCK,    "PM_FLOATING_EYE"},
{S_KOP,           "PM_KNIGHT"},
{S_LICH,          "PM_LICH"},
{S_MUMMY,         "PM_HUMAN_MUMMY"},
{S_NAGA,          "PM_GUARDIAN_NAGA"},
{S_OGRE,          "PM_OGRE"},
{S_PUDDING,       "PM_GREEN_SLIME"},
{S_SNAKE,         "PM_WATER_MOCCASIN"},
{S_TROLL,         "PM_TROLL"},
{S_WRAITH,        "PM_WRAITH"},
{S_ZOMBIE,        "PM_HUMAN_ZOMBIE"},
{S_GHOST,         "PM_GHOST"},
{S_GOLEM,         "PM_STONE_GOLEM"},
{S_DEMON,         "PM_WATER_DEMON"},
{S_LIZARD,        "PM_LIZARD"},
{S_VAMPIRE,       "PM_VAMPIRE"},
{S_WORM,          "PM_BABY_PURPLE_WORM"}
};



const struct eqmap tilemap[] = {
/*type       x      looks like     y*/
{T_MONSTER, "PM_SAMURAI",   "PM_KNIGHT" },
{T_MONSTER, "PM_MONK",      "PM_WIZARD" },
{T_MONSTER, "PM_HEALER",    "PM_PRIEST" },
{T_MONSTER, "PM_PRIESTESS", "PM_PRIEST" },
/* slashem only */
{T_MONSTER, "PM_NECROMANCER", "PM_WIZARD" },
{T_MONSTER, "PM_FLAME_MAGE",  "PM_WIZARD" },
{T_MONSTER, "PM_ICE_MAGE",    "PM_WIZARD" },

/* objects */
{T_OBJECT, "ELVEN_SHIELD",         "SMALL_SHIELD"},
{T_OBJECT, "URUK_HAI_SHIELD",      "SMALL_SHIELD"},
{T_OBJECT, "ORCISH_SHIELD",        "SMALL_SHIELD"},
{T_OBJECT, "LARGE_SHIELD",         "SMALL_SHIELD"},
{T_OBJECT, "DWARVISH_ROUNDSHIELD", "SMALL_SHIELD"},
{T_OBJECT, "SHIELD_OF_REFLECTION", "SMALL_SHIELD"},

{T_OBJECT, "ELVEN_LEATHER_HELM",         "HELMET"},
{T_OBJECT, "ORCISH_HELM",                "HELMET"},
{T_OBJECT, "DWARVISH_IRON_HELM",         "HELMET"},
{T_OBJECT, "DENTED_POT",                 "HELMET"},
{T_OBJECT, "HELM_OF_BRILLIANCE",         "HELMET"},
{T_OBJECT, "HELM_OF_OPPOSITE_ALIGNMENT", "HELMET"},
{T_OBJECT, "HELM_OF_TELEPATHY",          "HELMET"},

{T_OBJECT, "SHORT_SWORD",          "LONG_SWORD"},
{T_OBJECT, "ELVEN_SHORT_SWORD",    "LONG_SWORD"},
{T_OBJECT, "ORCISH_SHORT_SWORD",   "LONG_SWORD"},
{T_OBJECT, "DWARVISH_SHORT_SWORD", "LONG_SWORD"},
{T_OBJECT, "SCIMITAR",             "LONG_SWORD"},
{T_OBJECT, "SILVER_SABER",         "LONG_SWORD"},
{T_OBJECT, "BROADSWORD",           "LONG_SWORD"},
{T_OBJECT, "ELVEN_BROADSWORD",     "LONG_SWORD"},
{T_OBJECT, "TWO_HANDED_SWORD",     "LONG_SWORD"},
{T_OBJECT, "KATANA",               "LONG_SWORD"},
{T_OBJECT, "TSURUGI",              "LONG_SWORD"},
{T_OBJECT, "RUNESWORD",            "LONG_SWORD"},

{T_OBJECT, "ELVEN_SPEAR",    "SPEAR"},
{T_OBJECT, "ORCISH_SPEAR",   "SPEAR"},
{T_OBJECT, "DWARVISH_SPEAR", "SPEAR"},
{T_OBJECT, "SILVER_SPEAR",   "SPEAR"},
{T_OBJECT, "JAVELIN",        "SPEAR"},
{T_OBJECT, "TRIDENT",        "SPEAR"},

{T_OBJECT, "ELVEN_DAGGER",  "DAGGER"},
{T_OBJECT, "ORCISH_DAGGER", "DAGGER"},
{T_OBJECT, "SILVER_DAGGER", "DAGGER"},
{T_OBJECT, "ATHAME",        "DAGGER"},
{T_OBJECT, "SCALPEL",       "DAGGER"},
{T_OBJECT, "KNIFE",         "DAGGER"},
{T_OBJECT, "STILETTO",      "DAGGER"},
{T_OBJECT, "WORM_TOOTH",    "DAGGER"},
{T_OBJECT, "CRYSKNIFE",     "DAGGER"},

{T_OBJECT, "BATTLE_AXE", "AXE"},

{T_OBJECT, "LUCERN_HAMMER", "WAR_HAMMER"},

{T_OBJECT, "IRON_SHOES",          "LOW_BOOTS"},
{T_OBJECT, "HIGH_BOOTS",          "LOW_BOOTS"},
{T_OBJECT, "SPEED_BOOTS",         "LOW_BOOTS"},
{T_OBJECT, "WATER_WALKING_BOOTS", "LOW_BOOTS"},
{T_OBJECT, "JUMPING_BOOTS",       "LOW_BOOTS"},
{T_OBJECT, "ELVEN_BOOTS",         "LOW_BOOTS"},
{T_OBJECT, "KICKING_BOOTS",       "LOW_BOOTS"},
{T_OBJECT, "FUMBLE_BOOTS",        "LOW_BOOTS"},
{T_OBJECT, "LEVITATION_BOOTS",    "LOW_BOOTS"},

{T_OBJECT, "MORNING_STAR", "MACE"},
{T_OBJECT, "FLAIL",        "MACE"},

{T_OBJECT, "STUDDED_LEATHER_ARMOR", "LEATHER_ARMOR"},
{T_OBJECT, "LEATHER_JACKET",        "LEATHER_ARMOR"},

{T_OBJECT, "GRAY_DRAGON_SCALE_MAIL",   "SCALE_MAIL"},
{T_OBJECT, "SILVER_DRAGON_SCALE_MAIL", "SCALE_MAIL"},
{T_OBJECT, "RED_DRAGON_SCALE_MAIL",    "SCALE_MAIL"},
{T_OBJECT, "WHITE_DRAGON_SCALE_MAIL",  "SCALE_MAIL"},
{T_OBJECT, "ORANGE_DRAGON_SCALE_MAIL", "SCALE_MAIL"},
{T_OBJECT, "BLACK_DRAGON_SCALE_MAIL",  "SCALE_MAIL"},
{T_OBJECT, "BLUE_DRAGON_SCALE_MAIL",   "SCALE_MAIL"},
{T_OBJECT, "GREEN_DRAGON_SCALE_MAIL",  "SCALE_MAIL"},
{T_OBJECT, "YELLOW_DRAGON_SCALE_MAIL", "SCALE_MAIL"},
/* next 2 are slashem only */
{T_OBJECT, "SHIMMERING_DRAGON_SCALE_MAIL","SCALE_MAIL"},
{T_OBJECT, "DEEP_DRAGON_SCALE_MAIL",      "SCALE_MAIL"},

{T_OBJECT, "CRYSTAL_PLATE_MAIL", "PLATE_MAIL"},
{T_OBJECT, "BRONZE_PLATE_MAIL",  "PLATE_MAIL"},
{T_OBJECT, "SPLINT_MAIL",        "PLATE_MAIL"},
{T_OBJECT, "BANDED_MAIL",        "PLATE_MAIL"},

{T_OBJECT, "ORCISH_CHAIN_MAIL",       "CHAIN_MAIL"},
{T_OBJECT, "ELVEN_MITHRIL_COAT",      "CHAIN_MAIL"},
{T_OBJECT, "DARK_ELVEN_MITHRIL_COAT", "CHAIN_MAIL"},
{T_OBJECT, "DWARVISH_MITHRIL_COAT",   "CHAIN_MAIL"},

{T_OBJECT, "ORCISH_RING_MAIL", "RING_MAIL"},

{T_OBJECT, "GAUNTLETS_OF_FUMBLING",  "LEATHER_GLOVES"},
{T_OBJECT, "GAUNTLETS_OF_POWER",     "LEATHER_GLOVES"},
{T_OBJECT, "GAUNTLETS_OF_SWIMMING",  "LEATHER_GLOVES"},
{T_OBJECT, "GAUNTLETS_OF_DEXTERITY", "LEATHER_GLOVES"},
{T_OBJECT, "GAUNTLETS_OF_POWER",     "LEATHER_GLOVES"},

{T_OBJECT, "BRASS_LANTERN", "OIL_LAMP"},
{T_OBJECT, "MAGIC_LAMP",    "OIL_LAMP"},

{T_OBJECT, "TALLOW_CANDLE", "WAX_CANDLE"},
{T_OBJECT, "MAGIC_CANDLE",  "WAX_CANDLE"}, /* slashem only */

{T_OBJECT, "ELVEN_ARROW",   "ARROW"},
{T_OBJECT, "ORCISH_ARROW",  "ARROW"},
{T_OBJECT, "SILVER_ARROW",  "ARROW"},
{T_OBJECT, "YA",            "ARROW"},
{T_OBJECT, "CROSSBOW_BOLT", "ARROW"},
{T_OBJECT, "DART",          "ARROW"},

{T_OBJECT, "ELVEN_BOW",  "BOW"},
{T_OBJECT, "ORCISH_BOW", "BOW"},
{T_OBJECT, "YUMI",       "BOW"},
{T_OBJECT, "SLING",      "BOW"},

{T_OBJECT, "MUMMY_WRAPPING",            "ELVEN_CLOAK"},
{T_OBJECT, "ORCISH_CLOAK",              "ELVEN_CLOAK"},
{T_OBJECT, "DWARVISH_CLOAK",            "ELVEN_CLOAK"},
{T_OBJECT, "OILSKIN_CLOAK",             "ELVEN_CLOAK"},
{T_OBJECT, "ROBE",                      "ELVEN_CLOAK"},
{T_OBJECT, "ALCHEMY_SMOCK",             "ELVEN_CLOAK"},
{T_OBJECT, "LEATHER_CLOAK",             "ELVEN_CLOAK"},
{T_OBJECT, "CLOAK_OF_PROTECTION",       "ELVEN_CLOAK"},
{T_OBJECT, "CLOAK_OF_INVISIBILITY",     "ELVEN_CLOAK"},
{T_OBJECT, "CLOAK_OF_MAGIC_RESISTANCE", "ELVEN_CLOAK"},
{T_OBJECT, "CLOAK_OF_DISPLACEMENT",     "ELVEN_CLOAK"},
{T_OBJECT, "POISONOUS_CLOAK",           "ELVEN_CLOAK"},
/* next 4 are slashem only */
{T_OBJECT, "ROBE_OF_PROTECTION",        "ELVEN_CLOAK"},
{T_OBJECT, "ROBE_OF_POWER",             "ELVEN_CLOAK"},
{T_OBJECT, "ROBE_OF_WEAKNESS",          "ELVEN_CLOAK"},
{T_OBJECT, "LAB_COAT",                  "ELVEN_CLOAK"},

{T_OBJECT, "DUNCE_CAP", "CORNUTHAUM"},

{T_OBJECT, "LOCK_PICK",  "SKELETON_KEY"},

{T_OBJECT, "GRAPPLING_HOOK", "PICK_AXE"},

{T_OBJECT, "BELL_OF_OPENING", "BELL"},

{T_OBJECT, "CAN_OF_GREASE", "TIN"},

{T_OBJECT, "MAGIC_FLUTE", "WOODEN_FLUTE"},
{T_OBJECT, "MAGIC_HARP", "WOODEN_HARP"},

{T_OBJECT, "DRUM_OF_EARTHQUAKE", "LEATHER_DRUM"},

{T_OBJECT, "HORN_OF_PLENTY", "FROST_HORN"},
};




const struct tiledef tiledefinitions[] = {
/* Special tiles: anything that does not directly correspond to a monster or object
 * NAME may be anything; it will return in the output file as VTILE_<NAME> */
/* special: NAME                   FILENAME          top left x coord      top left y coord    width       height*/
{T_SPECIAL,"NONE",                 NULL,             0,                    0,                  0,          0},
{T_SPECIAL,"PLAYER_INVIS",         NULL,             0,                    0,                  0,          0},
{T_SPECIAL,"WALL_GENERIC",         NULL,             0,                    0,                  0,          0},
/* floor placeholders: these are used until we determine the actual tile in the floor pattern */
{T_SPECIAL,"FLOOR_COBBLESTONE",    NULL,             0,                    0,                  0,          0},
{T_SPECIAL,"FLOOR_ROUGH",          NULL,             0,                    0,                  0,          0},
{T_SPECIAL,"FLOOR_WATER",          NULL,             0,                    0,                  0,          0},
{T_SPECIAL,"FLOOR_ICE",            NULL,             0,                    0,                  0,          0},
{T_SPECIAL,"FLOOR_LAVA",           NULL,             0,                    0,                  0,          0},
{T_SPECIAL,"FLOOR_ROUGH_LIT",      NULL,             0,                    0,                  0,          0},
{T_SPECIAL,"FLOOR_AIR",            NULL,             0,                    0,                  0,          0},
{T_SPECIAL,"FLOOR_DARK",           NULL,             0,                    0,                  0,          0},
{T_SPECIAL,"FLOOR_NOT_VISIBLE",    "jtp_cm07.png",   1+(3*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"DOOR_WOOD_BROKEN",     "jtp_cm07.png",   1+(0*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"HDOOR_WOOD_CLOSED",    "jtp_cm07.png",   1+(1*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"VDOOR_WOOD_CLOSED",    "jtp_cm07.png",   1+(2*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"VDOOR_WOOD_OPEN",      "jtp_cm07.png",   1+(3*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"HDOOR_WOOD_OPEN",      "jtp_cm07.png",   1+(4*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_BEAR",            "jtp_cm07.png",   1+(5*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"GRAVE",                "jtp_cm07.png",   1+(0*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"ALTAR",                "jtp_cm07.png",   1+(1*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"FOUNTAIN",             "jtp_cm07.png",   1+(2*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"STAIRS_UP",            "jtp_cm07.png",   1+(3*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"STAIRS_DOWN",          "jtp_cm07.png",   1+(4*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"SINK",                 "jtp_cm07.png",   1+(5*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"GAS_TRAP",             "jtp_cm07.png",   1+(0*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_PIT",             "jtp_cm07.png",   1+(1*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_POLYMORPH",       "jtp_cm07.png",   1+(2*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TREE",                 "jtp_cm07.png",   1+(3*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_MAGIC",           "jtp_cm07.png",   1+(5*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_DOOR",            "jtp_cm07.png",   1+(0*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_WATER",           "jtp_cm07.png",   1+(1*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_TELEPORTER",      "jtp_cm07.png",   1+(2*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"UNMAPPED_AREA",        "jtp_cm07.png",   1+(4*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"HILITE_PET",           "jtp_cm07.png",   1+(5*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"BARS",                 "jtp_cmc7.png",   1+(0*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"THRONE",               "jtp_cmc7.png",   1+(1*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"EXPLOSION_NORTHWEST",  "jtp_cmc7.png",   1+(3*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"EXPLOSION_NORTH",      "jtp_cmc7.png",   1+(4*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"EXPLOSION_NORTHEAST",  "jtp_cmc7.png",   1+(5*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_ANTI_MAGIC",      "jtp_cmc7.png",   1+(0*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_ARROW",           "jtp_cmc7.png",   1+(1*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"EXPLOSION_WEST",       "jtp_cmc7.png",   1+(3*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"EXPLOSION_CENTER",     "jtp_cmc7.png",   1+(4*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"EXPLOSION_EAST",       "jtp_cmc7.png",   1+(5*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_FIRE",            "jtp_cmc7.png",   1+(0*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"ROLLING_BOULDER_TRAP", "jtp_cmc7.png",   1+(1*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"TRAP_SLEEPGAS",        "jtp_cmc7.png",   1+(2*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"EXPLOSION_SOUTHWEST",  "jtp_cmc7.png",   1+(3*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"EXPLOSION_SOUTH",      "jtp_cmc7.png",   1+(4*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"EXPLOSION_SOUTHEAST",  "jtp_cmc7.png",   1+(5*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"ZAP_SLANT_RIGHT",      "jtp_cmc7.png",   1+(0*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"ZAP_SLANT_LEFT",       "jtp_cmc7.png",   1+(1*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"ZAP_HORIZONTAL",       "jtp_cmc7.png",   1+(2*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"ZAP_VERTICAL",         "jtp_cmc7.png",   1+(3*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"LADDER_UP",            "jtp_cmc7.png",   1+(4*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"LADDER_DOWN",          "jtp_cmc7.png",   1+(5*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"RESIST_SPELL_1",       "jtp_cmc1.png",   1+(0*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"RESIST_SPELL_2",       "jtp_cmc1.png",   1+(1*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"RESIST_SPELL_3",       "jtp_cmc1.png",   1+(2*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"RESIST_SPELL_4",       "jtp_cmc1.png",   1+(3*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"WEB_TRAP",             "jtp_cmc1.png",   1+(4*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"DART_TRAP",            "jtp_cmc1.png",   1+(5*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"FALLING_ROCK_TRAP",    "jtp_cmc1.png",   1+(0*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"SQUEAKY_BOARD",        "jtp_cmc1.png",   1+(1*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"MAGIC_PORTAL",         "jtp_cmc1.png",   1+(3*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"SPIKED_PIT",           "jtp_cmc1.png",   1+(4*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"HOLE",                 "jtp_cmc1.png",   1+(5*CMAP_DX),        1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"LEVEL_TELEPORTER",     "jtp_cmc1.png",   1+(0*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"MAGIC_TRAP",           "jtp_cmc1.png",   1+(1*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"DIGBEAM",              "jtp_cmc1.png",   1+(2*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"FLASHBEAM",            "jtp_cmc1.png",   1+(3*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"BOOMLEFT",             "jtp_cmc1.png",   1+(4*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"BOOMRIGHT",            "jtp_cmc1.png",   1+(5*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"HCDBRIDGE",            "jtp_cmc1.png",   1+(0*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"VCDBRIDGE",            "jtp_cmc1.png",   1+(1*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"VODBRIDGE",            "jtp_cmc1.png",   1+(2*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"HODBRIDGE",            "jtp_cmc1.png",   1+(3*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"CLOUD",                "jtp_cmc1.png",   1+(4*CMAP_DX),        1+(3*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"STONE",                "jtp_obj2.png",   1+(3*OBJ_DX),         1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_SPECIAL,"HOLY_WATER",           "jtp_obj4.png",   1+(2*OBJ_DX),         1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_SPECIAL,"UNHOLY_WATER",         "jtp_obj4.png",   1+(3*OBJ_DX),         1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_SPECIAL,"UNIDENTIFIED_BAG",     "jtp_obj4.png",   1+(2*OBJ_DX),         1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_SPECIAL,"LAND_MINE",            "jtp_obj5.png",   1+(7*OBJ_DX),         1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_SPECIAL,"LAWFUL_PRIEST",        "jtp_mon1.png",   1+(5*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"CHAOTIC_PRIEST",       "jtp_mon1.png",   1+(0*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"NEUTRAL_PRIEST",       "jtp_mon1.png",   1+(1*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"UNALIGNED_PRIEST",     "jtp_mon1.png",   1+(2*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_A",        "jtp_monmisc1.png",1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_B",        "jtp_monmisc1.png",1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_C",        "jtp_monmisc1.png",1+(2*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_D",        "jtp_monmisc1.png",1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_E",        "jtp_monmisc1.png",1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_F",        "jtp_monmisc1.png",1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_G",        "jtp_monmisc1.png",1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_H",        "jtp_monmisc1.png",1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_I",        "jtp_monmisc1.png",1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_J",        "jtp_monmisc1.png",1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_K",        "jtp_monmisc1.png",1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_L",        "jtp_monmisc1.png",1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_M",        "jtp_monmisc1.png",1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_N",        "jtp_monmisc1.png",1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_O",        "jtp_monmisc1.png",1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_P",        "jtp_monmisc1.png",1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_Q",        "jtp_monmisc1.png",1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_R",        "jtp_monmisc1.png",1+(5*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_S",        "jtp_monmisc1.png",1+(0*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_T",        "jtp_monmisc1.png",1+(1*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_U",        "jtp_monmisc1.png",1+(2*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_V",        "jtp_monmisc1.png",1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_W",        "jtp_monmisc1.png",1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_X",        "jtp_monmisc1.png",1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_Y",        "jtp_monmisc2.png",1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ROGUE_LEVEL_Z",        "jtp_monmisc2.png",1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_SPECIAL,"ENGULF_FIRE_VORTEX",   "engulf.png",     1+(0*218),            1+(0*234),          214,        230},
{T_SPECIAL,"ENGULF_FOG_CLOUD",     "engulf.png",     1+(1*218),            1+(0*234),          214,        230},
{T_SPECIAL,"ENGULF_AIR_ELEMENTAL", "engulf.png",     1+(2*218),            1+(0*234),          214,        230},
{T_SPECIAL,"ENGULF_STEAM_VORTEX",  "engulf.png",     1+(0*218),            1+(1*234),          214,        230},
{T_SPECIAL,"ENGULF_PURPLE_WORM",   "engulf.png",     1+(1*218),            1+(1*234),          214,        230},
{T_SPECIAL,"ENGULF_JUIBLEX",       "engulf.png",     1+(2*218),            1+(1*234),          214,        230},
{T_SPECIAL,"ENGULF_OCHRE_JELLY",   "engulf2.png",    1+(0*218),            1+(0*234),          214,        230},
{T_SPECIAL,"ENGULF_LURKER_ABOVE",  "engulf2.png",    1+(1*218),            1+(0*234),          214,        230},
{T_SPECIAL,"ENGULF_TRAPPER",       "engulf2.png",    1+(2*218),            1+(0*234),          214,        230},
{T_SPECIAL,"ENGULF_DUST_VORTEX",   "engulf2.png",    1+(0*218),            1+(1*234),          214,        230},
{T_SPECIAL,"ENGULF_ICE_VORTEX",    "engulf2.png",    1+(1*218),            1+(1*234),          214,        230},
{T_SPECIAL,"ENGULF_ENERGY_VORTEX", "engulf2.png",    1+(2*218),            1+(1*234),          214,        230},
{T_SPECIAL,"WARNLEV_1",            "unknown.png",    1+(5*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"WARNLEV_2",            "unknown.png",    1+(4*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"WARNLEV_3",            "unknown.png",    1+(3*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"WARNLEV_4",            "unknown.png",    1+(2*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"WARNLEV_5",            "unknown.png",    1+(1*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"WARNLEV_6",            "unknown.png",    1+(0*CMAP_DX),        1+(0*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_SPECIAL,"INVISIBLE_MONSTER",    "unknown.png",    1+(3*CMAP_DX),        1+(2*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},


/* Dungeon tiles: Walls, floors, floor edges
 * NAME is always a basename + extensions; it is used to find what group a tile belongs to but it is not output */
/* dungeon tiles:NAME               FILENAME          top left x coord      top left y coord    width       height*/
/* floor edges: name = basename + "_" + 1-12 */
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_1", "jtp_cm12.png",1,                   478,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_2", "jtp_cm12.png",117,                 478,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_3", "jtp_cm12.png",1,                   531,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_4", "jtp_cm12.png",117,                 531,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_5", "jtp_cm12.png",233,                 478,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_6", "jtp_cm12.png",349,                 478,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_7", "jtp_cm12.png",233,                 531,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_8", "jtp_cm12.png",349,                 531,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_9", "jtp_cm12.png",465,                 478,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_10","jtp_cm12.png",581,                 478,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_11","jtp_cm12.png",465,                 531,                112,        49},
{T_DUNGEON,"FLOOR_EDGE_COBBLESTONE_12","jtp_cm12.png",581,                 531,                112,        49},
/* floors: name = patternname + "_" + x-coord + "_" + y-coord */
{T_DUNGEON,"FLOOR_COBBLESTONE_0_0","jtp_cm12.png",   1,                    1,                  112,        49},
{T_DUNGEON,"FLOOR_COBBLESTONE_1_0","jtp_cm12.png",   117,                  1,                  112,        49},
{T_DUNGEON,"FLOOR_COBBLESTONE_2_0","jtp_cm12.png",   233,                  1,                  112,        49},
{T_DUNGEON,"FLOOR_COBBLESTONE_0_1","jtp_cm12.png",   1,                    54,                 112,        49},
{T_DUNGEON,"FLOOR_COBBLESTONE_1_1","jtp_cm12.png",   117,                  54,                 112,        49},
{T_DUNGEON,"FLOOR_COBBLESTONE_2_1","jtp_cm12.png",   233,                  54,                 112,        49},
{T_DUNGEON,"FLOOR_COBBLESTONE_0_2","jtp_cm12.png",   1,                    107,                112,        49},
{T_DUNGEON,"FLOOR_COBBLESTONE_1_2","jtp_cm12.png",   117,                  107,                112,        49},
{T_DUNGEON,"FLOOR_COBBLESTONE_2_2","jtp_cm12.png",   233,                  107,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_0_0",      "jtp_cm12.png",   1,                    160,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_1_0",      "jtp_cm12.png",   117,                  160,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_2_0",      "jtp_cm12.png",   233,                  160,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_0_1",      "jtp_cm12.png",   1,                    213,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_1_1",      "jtp_cm12.png",   117,                  213,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_2_1",      "jtp_cm12.png",   233,                  213,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_0_2",      "jtp_cm12.png",   1,                    266,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_1_2",      "jtp_cm12.png",   117,                  266,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_2_2",      "jtp_cm12.png",   233,                  266,                112,        49},
{T_DUNGEON,"FLOOR_CERAMIC_0_0",    "jtp_cm12.png",   1,                    319,                112,        49},
{T_DUNGEON,"FLOOR_CERAMIC_1_0",    "jtp_cm12.png",   117,                  319,                112,        49},
{T_DUNGEON,"FLOOR_CERAMIC_2_0",    "jtp_cm12.png",   233,                  319,                112,        49},
{T_DUNGEON,"FLOOR_CERAMIC_0_1",    "jtp_cm12.png",   1,                    372,                112,        49},
{T_DUNGEON,"FLOOR_CERAMIC_1_1",    "jtp_cm12.png",   117,                  372,                112,        49},
{T_DUNGEON,"FLOOR_CERAMIC_2_1",    "jtp_cm12.png",   233,                  372,                112,        49},
{T_DUNGEON,"FLOOR_CERAMIC_0_2",    "jtp_cm12.png",   1,                    425,                112,        49},
{T_DUNGEON,"FLOOR_CERAMIC_1_2",    "jtp_cm12.png",   117,                  425,                112,        49},
{T_DUNGEON,"FLOOR_CERAMIC_2_2",    "jtp_cm12.png",   233,                  425,                112,        49},
{T_DUNGEON,"FLOOR_LAVA_0_0",       "jtp_cm12.png",   349,                  1,                  112,        49},
{T_DUNGEON,"FLOOR_LAVA_1_0",       "jtp_cm12.png",   465,                  1,                  112,        49},
{T_DUNGEON,"FLOOR_LAVA_2_0",       "jtp_cm12.png",   581,                  1,                  112,        49},
{T_DUNGEON,"FLOOR_LAVA_0_1",       "jtp_cm12.png",   349,                  54,                 112,        49},
{T_DUNGEON,"FLOOR_LAVA_1_1",       "jtp_cm12.png",   465,                  54,                 112,        49},
{T_DUNGEON,"FLOOR_LAVA_2_1",       "jtp_cm12.png",   581,                  54,                 112,        49},
{T_DUNGEON,"FLOOR_LAVA_0_2",       "jtp_cm12.png",   349,                  107,                112,        49},
{T_DUNGEON,"FLOOR_LAVA_1_2",       "jtp_cm12.png",   465,                  107,                112,        49},
{T_DUNGEON,"FLOOR_LAVA_2_2",       "jtp_cm12.png",   581,                  107,                112,        49},
{T_DUNGEON,"FLOOR_WATER_0_0",      "jtp_cm12.png",   349,                  160,                112,        49},
{T_DUNGEON,"FLOOR_WATER_1_0",      "jtp_cm12.png",   465,                  160,                112,        49},
{T_DUNGEON,"FLOOR_WATER_2_0",      "jtp_cm12.png",   581,                  160,                112,        49},
{T_DUNGEON,"FLOOR_WATER_0_1",      "jtp_cm12.png",   349,                  213,                112,        49},
{T_DUNGEON,"FLOOR_WATER_1_1",      "jtp_cm12.png",   465,                  213,                112,        49},
{T_DUNGEON,"FLOOR_WATER_2_1",      "jtp_cm12.png",   581,                  213,                112,        49},
{T_DUNGEON,"FLOOR_WATER_0_2",      "jtp_cm12.png",   349,                  266,                112,        49},
{T_DUNGEON,"FLOOR_WATER_1_2",      "jtp_cm12.png",   465,                  266,                112,        49},
{T_DUNGEON,"FLOOR_WATER_2_2",      "jtp_cm12.png",   581,                  266,                112,        49},
{T_DUNGEON,"FLOOR_ICE_0_0",        "jtp_cm12.png",   349,                  319,                112,        49},
{T_DUNGEON,"FLOOR_ICE_1_0",        "jtp_cm12.png",   465,                  319,                112,        49},
{T_DUNGEON,"FLOOR_ICE_2_0",        "jtp_cm12.png",   581,                  319,                112,        49},
{T_DUNGEON,"FLOOR_ICE_0_1",        "jtp_cm12.png",   349,                  372,                112,        49},
{T_DUNGEON,"FLOOR_ICE_1_1",        "jtp_cm12.png",   465,                  372,                112,        49},
{T_DUNGEON,"FLOOR_ICE_2_1",        "jtp_cm12.png",   581,                  372,                112,        49},
{T_DUNGEON,"FLOOR_ICE_0_2",        "jtp_cm12.png",   349,                  425,                112,        49},
{T_DUNGEON,"FLOOR_ICE_1_2",        "jtp_cm12.png",   465,                  425,                112,        49},
{T_DUNGEON,"FLOOR_ICE_2_2",        "jtp_cm12.png",   581,                  425,                112,        49},
{T_DUNGEON,"FLOOR_MURAL_0_0",      "jtp_cm13.png",   1,                    1,                  112,        49},
{T_DUNGEON,"FLOOR_MURAL_1_0",      "jtp_cm13.png",   117,                  1,                  112,        49},
{T_DUNGEON,"FLOOR_MURAL_2_0",      "jtp_cm13.png",   233,                  1,                  112,        49},
{T_DUNGEON,"FLOOR_MURAL_0_1",      "jtp_cm13.png",   1,                    54,                 112,        49},
{T_DUNGEON,"FLOOR_MURAL_1_1",      "jtp_cm13.png",   117,                  54,                 112,        49},
{T_DUNGEON,"FLOOR_MURAL_2_1",      "jtp_cm13.png",   233,                  54,                 112,        49},
{T_DUNGEON, "FLOOR_MURAL2_0_0",    "jtp_cm13.png",   349,                  1,                  112,        49},
{T_DUNGEON, "FLOOR_MURAL2_1_0",    "jtp_cm13.png",   465,                  1,                  112,        49},
{T_DUNGEON, "FLOOR_MURAL2_2_0",    "jtp_cm13.png",   581,                  1,                  112,        49},
{T_DUNGEON, "FLOOR_MURAL2_0_1",    "jtp_cm13.png",   349,                  54,                 112,        49},
{T_DUNGEON, "FLOOR_MURAL2_1_1",    "jtp_cm13.png",   465,                  54,                 112,        49},
{T_DUNGEON, "FLOOR_MURAL2_2_1",    "jtp_cm13.png",   581,                  54,                 112,        49},
{T_DUNGEON,"FLOOR_CARPET_0_0",     "jtp_cm13.png",   1,                    107,                112,        49},
{T_DUNGEON,"FLOOR_CARPET_1_0",     "jtp_cm13.png",   117,                  107,                112,        49},
{T_DUNGEON,"FLOOR_CARPET_2_0",     "jtp_cm13.png",   233,                  107,                112,        49},
{T_DUNGEON,"FLOOR_CARPET_0_1",     "jtp_cm13.png",   1,                    160,                112,        49},
{T_DUNGEON,"FLOOR_CARPET_1_1",     "jtp_cm13.png",   117,                  160,                112,        49},
{T_DUNGEON,"FLOOR_CARPET_2_1",     "jtp_cm13.png",   233,                  160,                112,        49},
{T_DUNGEON,"FLOOR_MOSS_COVERED_0_0","jtp_cm13.png",  1,                    213,                112,        49},
{T_DUNGEON,"FLOOR_MOSS_COVERED_1_0","jtp_cm13.png",  117,                  213,                112,        49},
{T_DUNGEON,"FLOOR_MOSS_COVERED_2_0","jtp_cm13.png",  233,                  213,                112,        49},
{T_DUNGEON,"FLOOR_MOSS_COVERED_0_1","jtp_cm13.png",  1,                    266,                112,        49},
{T_DUNGEON,"FLOOR_MOSS_COVERED_1_1","jtp_cm13.png",  117,                  266,                112,        49},
{T_DUNGEON,"FLOOR_MOSS_COVERED_2_1","jtp_cm13.png",  233,                  266,                112,        49},
{T_DUNGEON,"FLOOR_MOSS_COVERED_0_2","jtp_cm13.png",  1,                    319,                112,        49},
{T_DUNGEON,"FLOOR_MOSS_COVERED_1_2","jtp_cm13.png",  117,                  319,                112,        49},
{T_DUNGEON,"FLOOR_MOSS_COVERED_2_2","jtp_cm13.png",  233,                  319,                112,        49},
{T_DUNGEON,"FLOOR_MARBLE_0_0",     "jtp_cm13.png",   349,                  213,                112,        49},
{T_DUNGEON,"FLOOR_MARBLE_1_0",     "jtp_cm13.png",   465,                  213,                112,        49},
{T_DUNGEON,"FLOOR_MARBLE_2_0",     "jtp_cm13.png",   581,                  213,                112,        49},
{T_DUNGEON,"FLOOR_MARBLE_0_1",     "jtp_cm13.png",   349,                  266,                112,        49},
{T_DUNGEON,"FLOOR_MARBLE_1_1",     "jtp_cm13.png",   465,                  266,                112,        49},
{T_DUNGEON,"FLOOR_MARBLE_2_1",     "jtp_cm13.png",   581,                  266,                112,        49},
{T_DUNGEON,"FLOOR_MARBLE_0_2",     "jtp_cm13.png",   349,                  319,                112,        49},
{T_DUNGEON,"FLOOR_MARBLE_1_2",     "jtp_cm13.png",   465,                  319,                112,        49},
{T_DUNGEON,"FLOOR_MARBLE_2_2",     "jtp_cm13.png",   581,                  319,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_LIT_0_0",  "jtp_cm13.png",   1,                    372,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_LIT_1_0",  "jtp_cm13.png",   117,                  372,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_LIT_2_0",  "jtp_cm13.png",   233,                  372,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_LIT_0_1",  "jtp_cm13.png",   1,                    425,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_LIT_1_1",  "jtp_cm13.png",   117,                  425,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_LIT_2_1",  "jtp_cm13.png",   233,                  425,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_LIT_0_2",  "jtp_cm13.png",   1,                    478,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_LIT_1_2",  "jtp_cm13.png",   117,                  478,                112,        49},
{T_DUNGEON,"FLOOR_ROUGH_LIT_2_2",  "jtp_cm13.png",   233,                  478,                112,        49},
{T_DUNGEON,"FLOOR_AIR_0_0",        "jtp_cm13.png",   349,                  372,                112,        49},
{T_DUNGEON,"FLOOR_AIR_1_0",        "jtp_cm13.png",   465,                  372,                112,        49},
{T_DUNGEON,"FLOOR_AIR_2_0",        "jtp_cm13.png",   581,                  372,                112,        49},
{T_DUNGEON,"FLOOR_AIR_0_1",        "jtp_cm13.png",   349,                  425,                112,        49},
{T_DUNGEON,"FLOOR_AIR_1_1",        "jtp_cm13.png",   465,                  425,                112,        49},
{T_DUNGEON,"FLOOR_AIR_2_1",        "jtp_cm13.png",   581,                  425,                112,        49},
{T_DUNGEON,"FLOOR_AIR_0_2",        "jtp_cm13.png",   349,                  478,                112,        49},
{T_DUNGEON,"FLOOR_AIR_1_2",        "jtp_cm13.png",   465,                  478,                112,        49},
{T_DUNGEON,"FLOOR_AIR_2_2",        "jtp_cm13.png",   581,                  478,                112,        49},
{T_DUNGEON,"FLOOR_DARK_0_0",       "jtp_cm13.png",   1,                    531,                112,        49},
/* walls: name = basename + "_" + (F|H|T) + "_" + (W|N|S|E) 
 * where (F|H|T) stands for full height/half height/transparent and (W|N|S|E) is west/north/south/east */
{T_DUNGEON,"WALL_BRICK_F_W",       "walls02a.png",   1,                    1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_F_N",       "walls02a.png",   61,                   1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_F_S",       "walls02a.png",   121,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_F_E",       "walls02a.png",   181,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_F_W","walls02a.png",   241,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_F_N","walls02a.png",   301,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_F_S","walls02a.png",   361,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_F_E","walls02a.png",   421,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_F_W","walls02a.png", 1,                    127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_F_N","walls02a.png", 61,                   127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_F_S","walls02a.png", 121,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_F_E","walls02a.png", 181,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_F_W","walls02a.png",   241,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_F_N","walls02a.png",   301,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_F_S","walls02a.png",   361,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_F_E","walls02a.png",   421,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_F_W","walls02a.png",   481,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_F_N","walls02a.png",   541,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_F_S","walls02a.png",   601,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_F_E","walls02a.png",   661,                  1,                  56,         122},
{T_DUNGEON,"WALL_DARK_F_W",        "walls04a.png",   1,                    1,                  56,         122},
{T_DUNGEON,"WALL_DARK_F_N",        "walls04a.png",   61,                   1,                  56,         122},
{T_DUNGEON,"WALL_DARK_F_S",        "walls04a.png",   121,                  1,                  56,         122},
{T_DUNGEON,"WALL_DARK_F_E",        "walls04a.png",   181,                  1,                  56,         122},
{T_DUNGEON,"WALL_LIGHT_F_W",       "walls04a.png",   1,                    127,                56,         122},
{T_DUNGEON,"WALL_LIGHT_F_N",       "walls04a.png",   61,                   127,                56,         122},
{T_DUNGEON,"WALL_LIGHT_F_S",       "walls04a.png",   121,                  127,                56,         122},
{T_DUNGEON,"WALL_LIGHT_F_E",       "walls04a.png",   181,                  127,                56,         122},
{T_DUNGEON,"WALL_MARBLE_F_W",      "walls02a.png",   1,                    253,                56,         122},
{T_DUNGEON,"WALL_MARBLE_F_N",      "walls02a.png",   61,                   253,                56,         122},
{T_DUNGEON,"WALL_MARBLE_F_S",      "walls02a.png",   121,                  253,                56,         122},
{T_DUNGEON,"WALL_MARBLE_F_E",      "walls02a.png",   181,                  253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_F_W",       "walls03a.png",   1,                    253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_F_N",       "walls03a.png",   61,                   253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_F_S",       "walls03a.png",   121,                  253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_F_E",       "walls03a.png",   181,                  253,                56,         122},
{T_DUNGEON,"WALL_STUCCO_F_W",      "walls03a.png",   1,                    127,                56,         122},
{T_DUNGEON,"WALL_STUCCO_F_N",      "walls03a.png",   61,                   127,                56,         122},
{T_DUNGEON,"WALL_STUCCO_F_S",      "walls03a.png",   121,                  127,                56,         122},
{T_DUNGEON,"WALL_STUCCO_F_E",      "walls03a.png",   181,                  127,                56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_F_W","walls03a.png",   1,                    1,                  56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_F_N","walls03a.png",   61,                   1,                  56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_F_S","walls03a.png",   121,                  1,                  56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_F_E","walls03a.png",   181,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_H_W",       "walls02c.png",   1,                    1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_H_N",       "walls02c.png",   61,                   1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_H_S",       "walls02c.png",   121,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_H_E",       "walls02c.png",   181,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_H_W","walls02c.png",   241,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_H_N","walls02c.png",   301,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_H_S","walls02c.png",   361,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_H_E","walls02c.png",   421,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_H_W","walls02c.png", 1,                    127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_H_N","walls02c.png", 61,                   127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_H_S","walls02c.png", 121,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_H_E","walls02c.png", 181,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_H_W","walls02c.png",   241,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_H_N","walls02c.png",   301,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_H_S","walls02c.png",   361,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_H_E","walls02c.png",   421,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_H_W","walls02c.png",   481,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_H_N","walls02c.png",   541,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_H_S","walls02c.png",   601,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_H_E","walls02c.png",   661,                  1,                  56,         122},
{T_DUNGEON,"WALL_DARK_H_W",        "walls04c.png",   1,                    1,                  56,         122},
{T_DUNGEON,"WALL_DARK_H_N",        "walls04c.png",   61,                   1,                  56,         122},
{T_DUNGEON,"WALL_DARK_H_S",        "walls04c.png",   121,                  1,                  56,         122},
{T_DUNGEON,"WALL_DARK_H_E",        "walls04c.png",   181,                  1,                  56,         122},
{T_DUNGEON,"WALL_LIGHT_H_W",       "walls04c.png",   1,                    127,                56,         122},
{T_DUNGEON,"WALL_LIGHT_H_N",       "walls04c.png",   61,                   127,                56,         122},
{T_DUNGEON,"WALL_LIGHT_H_S",       "walls04c.png",   121,                  127,                56,         122},
{T_DUNGEON,"WALL_LIGHT_H_E",       "walls04c.png",   181,                  127,                56,         122},
{T_DUNGEON,"WALL_MARBLE_H_W",      "walls02c.png",   1,                    253,                56,         122},
{T_DUNGEON,"WALL_MARBLE_H_N",      "walls02c.png",   61,                   253,                56,         122},
{T_DUNGEON,"WALL_MARBLE_H_S",      "walls02c.png",   121,                  253,                56,         122},
{T_DUNGEON,"WALL_MARBLE_H_E",      "walls02c.png",   181,                  253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_H_W",       "walls03c.png",   1,                    253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_H_N",       "walls03c.png",   61,                   253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_H_S",       "walls03c.png",   121,                  253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_H_E",       "walls03c.png",   181,                  253,                56,         122},
{T_DUNGEON,"WALL_STUCCO_H_W",      "walls03c.png",   1,                    127,                56,         122},
{T_DUNGEON,"WALL_STUCCO_H_N",      "walls03c.png",   61,                   127,                56,         122},
{T_DUNGEON,"WALL_STUCCO_H_S",      "walls03c.png",   121,                  127,                56,         122},
{T_DUNGEON,"WALL_STUCCO_H_E",      "walls03c.png",   181,                  127,                56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_H_W","walls03c.png",   1,                    1,                  56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_H_N","walls03c.png",   61,                   1,                  56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_H_S","walls03c.png",   121,                  1,                  56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_H_E","walls03c.png",   181,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_T_W",       "walls02b.png",   1,                    1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_T_N",       "walls02b.png",   61,                   1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_T_S",       "walls02b.png",   121,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_T_E",       "walls02b.png",   181,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_T_W","walls02b.png",   241,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_T_N","walls02b.png",   301,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_T_S","walls02b.png",   361,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_BANNER_T_E","walls02b.png",   421,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_T_W","walls02b.png", 1,                    127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_T_N","walls02b.png", 61,                   127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_T_S","walls02b.png", 121,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PAINTING_T_E","walls02b.png", 181,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_T_W","walls02b.png",   241,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_T_N","walls02b.png",   301,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_T_S","walls02b.png",   361,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_PILLAR_T_E","walls02b.png",   421,                  127,                56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_T_W","walls02b.png",   481,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_T_N","walls02b.png",   541,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_T_S","walls02b.png",   601,                  1,                  56,         122},
{T_DUNGEON,"WALL_BRICK_POCKET_T_E","walls02b.png",   661,                  1,                  56,         122},
{T_DUNGEON,"WALL_DARK_T_W",        "walls04b.png",   1,                    1,                  56,         122},
{T_DUNGEON,"WALL_DARK_T_N",        "walls04b.png",   61,                   1,                  56,         122},
{T_DUNGEON,"WALL_DARK_T_S",        "walls04b.png",   121,                  1,                  56,         122},
{T_DUNGEON,"WALL_DARK_T_E",        "walls04b.png",   181,                  1,                  56,         122},
{T_DUNGEON,"WALL_LIGHT_T_W",       "walls04b.png",   1,                    127,                56,         122},
{T_DUNGEON,"WALL_LIGHT_T_N",       "walls04b.png",   61,                   127,                56,         122},
{T_DUNGEON,"WALL_LIGHT_T_S",       "walls04b.png",   121,                  127,                56,         122},
{T_DUNGEON,"WALL_LIGHT_T_E",       "walls04b.png",   181,                  127,                56,         122},
{T_DUNGEON,"WALL_MARBLE_T_W",      "walls02b.png",   1,                    253,                56,         122},
{T_DUNGEON,"WALL_MARBLE_T_N",      "walls02b.png",   61,                   253,                56,         122},
{T_DUNGEON,"WALL_MARBLE_T_S",      "walls02b.png",   121,                  253,                56,         122},
{T_DUNGEON,"WALL_MARBLE_T_E",      "walls02b.png",   181,                  253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_T_W",       "walls03b.png",   1,                    253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_T_N",       "walls03b.png",   61,                   253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_T_S",       "walls03b.png",   121,                  253,                56,         122},
{T_DUNGEON,"WALL_ROUGH_T_E",       "walls03b.png",   181,                  253,                56,         122},
{T_DUNGEON,"WALL_STUCCO_T_W",      "walls03b.png",   1,                    127,                56,         122},
{T_DUNGEON,"WALL_STUCCO_T_N",      "walls03b.png",   61,                   127,                56,         122},
{T_DUNGEON,"WALL_STUCCO_T_S",      "walls03b.png",   121,                  127,                56,         122},
{T_DUNGEON,"WALL_STUCCO_T_E",      "walls03b.png",   181,                  127,                56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_T_W","walls03b.png",   1,                    1,                  56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_T_N","walls03b.png",   61,                   1,                  56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_T_S","walls03b.png",   121,                  1,                  56,         122},
{T_DUNGEON,"WALL_VINE_COVERED_T_E","walls03b.png",   181,                  1,                  56,         122},



/* objects: NAME                   FILENAME          top left x coord     top left y coord    width        height
 * The name field is the name of a define in onames.h, _as a string_ !
 * It's okay for an object to be assigned to a tile that matches its description
 *(ie a potion of oil is assigned to a picture of a murky potion) because we shuffle those later */
{T_OBJECT, "LAND_MINE",            "jtp_cmc1.png",   1+(2*CMAP_DX),       1+(1*CMAP_DY),      CMAP_DX-4,  CMAP_DY-4},
{T_OBJECT, "STRANGE_OBJECT",       "jtp_obj1.png",   1+(0*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SKELETON_KEY",         "jtp_obj1.png",   1+(7*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BOULDER",              "jtp_obj1.png",   1+(1*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CORPSE",               "jtp_obj1.png",   1+(2*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BOW",                  "jtp_obj1.png",   1+(6*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "STATUE",               "jtp_obj1.png",   1+(0*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CHEST",                "jtp_obj1.png",   1+(1*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GOLD_PIECE",           "jtp_obj1.png",   1+(2*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SPE_DIG",              "jtp_obj1.png",   1+(3*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "HELMET",               "jtp_obj1.png",   1+(4*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SMALL_SHIELD",         "jtp_obj1.png",   1+(5*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "AMULET_OF_ESP",        "jtp_obj1.png",   1+(6*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "DAGGER",               "jtp_obj1.png",   1+(7*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LOW_BOOTS",            "jtp_obj1.png",   1+(0*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SPEAR",                "jtp_obj1.png",   1+(1*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SCR_ENCHANT_ARMOR",    "jtp_obj1.png",   1+(3*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "WAN_LIGHT",            "jtp_obj1.png",   1+(4*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LONG_SWORD",           "jtp_obj1.png",   1+(5*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "RIN_ADORNMENT",        "jtp_obj1.png",   1+(6*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "APPLE",                "jtp_obj1.png",   1+(7*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GEM_BLUE_GLASS",       "jtp_obj1.png",   1+(0*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "RING_MAIL",            "jtp_obj1.png",   1+(1*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LEATHER_ARMOR",        "jtp_obj1.png",   1+(2*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "PLATE_MAIL",           "jtp_obj1.png",   1+(3*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "WAR_HAMMER",           "jtp_obj1.png",   1+(4*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "AXE",                  "jtp_obj1.png",   1+(5*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "OIL_LAMP",             "jtp_obj1.png",   1+(6*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "PEAR",                 "jtp_obj1.png",   1+(7*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SCALE_MAIL",           "jtp_obj1.png",   1+(0*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CHAIN_MAIL",           "jtp_obj1.png",   1+(1*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "ELVEN_CLOAK",          "jtp_obj1.png",   1+(2*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TRIDENT",              "jtp_obj1.png",   1+(3*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "EXPENSIVE_CAMERA",     "jtp_obj1.png",   1+(4*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "FEDORA",               "jtp_obj1.png",   1+(5*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CLUB",                 "jtp_obj1.png",   1+(6*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "ARROW",                "jtp_obj1.png",   1+(7*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "EGG",                  "jtp_obj2.png",   1+(0*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LEATHER_GLOVES",       "jtp_obj2.png",   1+(1*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BELL",                 "jtp_obj2.png",   1+(2*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CANDY_BAR",            "jtp_obj2.png",   1+(6*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "WAX_CANDLE",           "jtp_obj2.png",   1+(5*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CARROT",               "jtp_obj2.png",   1+(7*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GEM_RED_GLASS",        "jtp_obj2.png",   1+(3*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GEM_GREEN_GLASS",      "jtp_obj2.png",   1+(4*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GEM_YELLOW_GLASS",     "jtp_obj2.png",   1+(0*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GEM_WHITE_GLASS",      "jtp_obj2.png",   1+(1*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GEM_BLACK_GLASS",      "jtp_obj2.png",   1+(2*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BULLWHIP",             "jtp_obj2.png",   1+(3*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "MACE",                 "jtp_obj2.png",   1+(4*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TOOLED_HORN",          "jtp_obj2.png",   1+(5*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "FORTUNE_COOKIE",       "jtp_obj2.png",   1+(6*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BANANA",               "jtp_obj2.png",   1+(7*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CRYSTAL_BALL",         "jtp_obj2.png",   1+(0*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "FROST_HORN",           "jtp_obj2.png",   1+(1*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "UNICORN_HORN",         "jtp_obj2.png",   1+(2*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "HAWAIIAN_SHIRT",       "jtp_obj2.png",   1+(3*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CREDIT_CARD",          "jtp_obj2.png",   1+(4*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "MIRROR",               "jtp_obj2.png",   1+(5*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CREAM_PIE",            "jtp_obj2.png",   1+(6*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "ORANGE",               "jtp_obj2.png",   1+(7*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CROSSBOW",             "jtp_obj2.png",   1+(0*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CORNUTHAUM",           "jtp_obj2.png",   1+(1*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "MAGIC_MARKER",         "jtp_obj2.png",   1+(2*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "QUARTERSTAFF",         "jtp_obj2.png",   1+(3*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "FOOD_RATION",          "jtp_obj2.png",   1+(4*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "PANCAKE",              "jtp_obj2.png",   1+(5*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LEMBAS_WAFER",         "jtp_obj2.png",   1+(6*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "KELP_FROND",           "jtp_obj2.png",   1+(7*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TRIPE_RATION",         "jtp_obj2.png",   1+(0*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "MEAT_STICK",           "jtp_obj2.png",   1+(1*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SHURIKEN",             "jtp_obj2.png",   1+(2*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "PICK_AXE",             "jtp_obj2.png",   1+(4*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TIN",                  "jtp_obj2.png",   1+(5*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CLOVE_OF_GARLIC",      "jtp_obj2.png",   1+(6*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "EUCALYPTUS_LEAF",      "jtp_obj2.png",   1+(7*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "MELON",                "jtp_obj3.png",   1+(0*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SLIME_MOLD",           "jtp_obj3.png",   1+(1*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SPRIG_OF_WOLFSBANE",   "jtp_obj3.png",   1+(2*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CRAM_RATION",          "jtp_obj3.png",   1+(3*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "HUGE_CHUNK_OF_MEAT",   "jtp_obj3.png",   1+(4*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "MEATBALL",             "jtp_obj3.png",   1+(5*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "K_RATION",             "jtp_obj3.png",   1+(6*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "C_RATION",             "jtp_obj3.png",   1+(7*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BOOMERANG",            "jtp_obj3.png",   1+(0*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LUMP_OF_ROYAL_JELLY",  "jtp_obj3.png",   1+(1*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "MEAT_RING",            "jtp_obj3.png",   1+(2*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "DILITHIUM_CRYSTAL",    "jtp_obj3.png",   1+(3*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "DIAMOND",              "jtp_obj3.png",   1+(4*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "RUBY",                 "jtp_obj3.png",   1+(5*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "JACINTH",              "jtp_obj3.png",   1+(6*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SAPPHIRE",             "jtp_obj3.png",   1+(7*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "EMERALD",              "jtp_obj3.png",   1+(0*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BLACK_OPAL",           "jtp_obj3.png",   1+(1*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TURQUOISE",            "jtp_obj3.png",   1+(2*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CITRINE",              "jtp_obj3.png",   1+(3*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "AQUAMARINE",           "jtp_obj3.png",   1+(4*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "AMBER",                "jtp_obj3.png",   1+(5*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TOPAZ",                "jtp_obj3.png",   1+(6*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "JET",                  "jtp_obj3.png",   1+(7*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "OPAL",                 "jtp_obj3.png",   1+(0*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CHRYSOBERYL",          "jtp_obj3.png",   1+(1*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GARNET",               "jtp_obj3.png",   1+(2*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "AMETHYST",             "jtp_obj3.png",   1+(3*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "JASPER",               "jtp_obj3.png",   1+(4*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "FLUORITE",             "jtp_obj3.png",   1+(5*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "OBSIDIAN",             "jtp_obj3.png",   1+(6*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "AGATE",                "jtp_obj3.png",   1+(7*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "JADE",                 "jtp_obj3.png",   1+(0*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LUCKSTONE",            "jtp_obj3.png",   1+(1*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LOADSTONE",            "jtp_obj3.png",   1+(2*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "FLINT",                "jtp_obj3.png",   1+(3*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GEM_VIOLET_GLASS",     "jtp_obj3.png",   1+(4*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GEM_ORANGE_GLASS",     "jtp_obj3.png",   1+(5*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TINNING_KIT",          "jtp_obj3.png",   1+(6*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "FIRE_HORN",            "jtp_obj3.png",   1+(7*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_GAIN_ABILITY",     "jtp_obj4.png",   1+(0*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_RESTORE_ABILITY",  "jtp_obj4.png",   1+(1*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_CONFUSION",        "jtp_obj4.png",   1+(2*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_BLINDNESS",        "jtp_obj4.png",   1+(3*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_PARALYSIS",        "jtp_obj4.png",   1+(4*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_SPEED",            "jtp_obj4.png",   1+(5*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_LEVITATION",       "jtp_obj4.png",   1+(6*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_HALLUCINATION",    "jtp_obj4.png",   1+(7*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_INVISIBILITY",     "jtp_obj4.png",   1+(0*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_SEE_INVISIBLE",    "jtp_obj4.png",   1+(1*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_HEALING",          "jtp_obj4.png",   1+(2*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_EXTRA_HEALING",    "jtp_obj4.png",   1+(3*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_GAIN_LEVEL",       "jtp_obj4.png",   1+(4*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_ENLIGHTENMENT",    "jtp_obj4.png",   1+(5*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_MONSTER_DETECTION","jtp_obj4.png",   1+(6*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_OBJECT_DETECTION", "jtp_obj4.png",   1+(7*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_GAIN_ENERGY",      "jtp_obj4.png",   1+(0*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_SLEEPING",         "jtp_obj4.png",   1+(1*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_FULL_HEALING",     "jtp_obj4.png",   1+(2*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_POLYMORPH",        "jtp_obj4.png",   1+(3*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_BOOZE",            "jtp_obj4.png",   1+(4*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_SICKNESS",         "jtp_obj4.png",   1+(5*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_FRUIT_JUICE",      "jtp_obj4.png",   1+(6*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_ACID",             "jtp_obj4.png",   1+(7*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_OIL",              "jtp_obj4.png",   1+(0*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "POT_WATER",            "jtp_obj4.png",   1+(1*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SACK",                 "jtp_obj4.png",   1+(4*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BAG_OF_HOLDING",       "jtp_obj4.png",   1+(5*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "OILSKIN_SACK",         "jtp_obj4.png",   1+(6*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BAG_OF_TRICKS",        "jtp_obj4.png",   1+(7*OBJ_DX),        1+(3*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "ICE_BOX",              "jtp_obj4.png",   1+(0*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LARGE_BOX",            "jtp_obj4.png",   1+(1*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TOUCHSTONE",           "jtp_obj4.png",   1+(3*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "ROCK",                 "jtp_obj4.png",   1+(4*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "HEALTHSTONE",          "jtp_obj4.png",   1+(6*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "WHETSTONE",            "jtp_obj4.png",   1+(7*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "GEM_BROWN_GLASS",      "jtp_obj4.png",   1+(5*OBJ_DX),        1+(4*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LEASH",                "jtp_obj5.png",   1+(0*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LENSES",               "jtp_obj5.png",   1+(1*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BLINDFOLD",            "jtp_obj5.png",   1+(2*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TOWEL",                "jtp_obj5.png",   1+(3*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "STETHOSCOPE",          "jtp_obj5.png",   1+(4*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "MAGIC_WHISTLE",        "jtp_obj5.png",   1+(5*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TIN_WHISTLE",          "jtp_obj5.png",   1+(6*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "SADDLE",               "jtp_obj5.png",   1+(7*OBJ_DX),        1+(0*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "WOODEN_FLUTE",         "jtp_obj5.png",   1+(0*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "WOODEN_HARP",          "jtp_obj5.png",   1+(1*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "CANDELABRUM_OF_INVOCATION","jtp_obj5.png",1+(2*OBJ_DX),       1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "IRON_CHAIN",           "jtp_obj5.png",   1+(3*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "HEAVY_IRON_BALL",      "jtp_obj5.png",   1+(4*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "LEATHER_DRUM",         "jtp_obj5.png",   1+(5*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "FIGURINE",             "jtp_obj5.png",   1+(6*OBJ_DX),        1+(1*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BEARTRAP",             "jtp_obj5.png",   1+(0*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "T_SHIRT",              "jtp_obj5.png",   1+(1*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "TIN_OPENER",           "jtp_obj5.png",   1+(2*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},
{T_OBJECT, "BUGLE",                "jtp_obj5.png",   1+(3*OBJ_DX),        1+(2*OBJ_DY),       OBJ_DX-4,   OBJ_DY-4},


/* figurines:NAME                   FILENAME          top left x coord     top left y coord    width       height*/
/* The name field is the name of a define in pm.h, _as a string_ ! */
{T_FIGURINE,"PM_FREEZING_SPHERE",   "figurine1.png",   1+(0*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_FLAMING_SPHERE",    "figurine1.png",   1+(1*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_SHOCKING_SPHERE",   "figurine1.png",   1+(2*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_PONY",              "figurine1.png",   1+(3*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_HORSE",             "figurine1.png",   1+(4*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WARHORSE",          "figurine1.png",   1+(5*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ACID_BLOB",         "figurine1.png",   1+(6*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_QUIVERING_BLOB",    "figurine1.png",   1+(7*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GELATINOUS_CUBE",   "figurine1.png",   1+(8*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_BROWN_MOLD",        "figurine1.png",   1+(9*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LICHEN",            "figurine1.png",   1+(10*FIG_DX),      1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GREEN_MOLD",        "figurine1.png",   1+(11*FIG_DX),      1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ROTHE",             "figurine1.png",   1+(0*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GNOME",             "figurine1.png",   1+(1*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GREEN_ELF",         "figurine1.png",   1+(2*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_DWARF",             "figurine1.png",   1+(3*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WATER_MOCCASIN",    "figurine1.png",   1+(4*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_MANES",             "figurine1.png",   1+(5*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_SHRIEKER",          "figurine1.png",   1+(6*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_STALKER",           "figurine1.png",   1+(7*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_AIR_ELEMENTAL",     "figurine1.png",   1+(8*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_EARTH_ELEMENTAL",   "figurine1.png",   1+(9*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WATER_ELEMENTAL",   "figurine1.png",   1+(10*FIG_DX),      1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_BLACK_LIGHT",       "figurine1.png",   1+(11*FIG_DX),      1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_HOBBIT",            "figurine1.png",   1+(0*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_BLACK_UNICORN",     "figurine1.png",   1+(1*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GRAY_UNICORN",      "figurine1.png",   1+(2*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WHITE_UNICORN",     "figurine1.png",   1+(3*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GIANT_ANT",         "figurine1.png",   1+(4*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_FIRE_ANT",          "figurine1.png",   1+(5*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ROCK_PIERCER",      "figurine1.png",   1+(6*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_IRON_PIERCER",      "figurine1.png",   1+(7*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GLASS_PIERCER",     "figurine1.png",   1+(8*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_DOG",               "figurine1.png",   1+(9*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LARGE_DOG",         "figurine1.png",   1+(10*FIG_DX),      1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WATER_DEMON",       "figurine1.png",   1+(11*FIG_DX),      1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ELVENKING",         "figurine1.png",   1+(0*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ELF_LORD",          "figurine1.png",   1+(1*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GREY_ELF",          "figurine1.png",   1+(2*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WOODLAND_ELF",      "figurine1.png",   1+(3*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ELF",               "figurine1.png",   1+(4*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ROCK_MOLE",         "figurine1.png",   1+(5*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GECKO",             "figurine1.png",   1+(6*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_NEWT",              "figurine1.png",   1+(7*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GAS_SPORE",         "figurine1.png",   1+(9*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_KITTEN",            "figurine1.png",   1+(10*FIG_DX),      1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LARGE_CAT",         "figurine1.png",   1+(11*FIG_DX),      1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_DWARF_LORD",        "figurine1.png",   1+(0*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_DWARF_KING",        "figurine1.png",   1+(1*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GNOME_LORD",        "figurine1.png",   1+(2*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GNOME_KING",        "figurine1.png",   1+(3*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GNOMISH_WIZARD",    "figurine1.png",   1+(4*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ORC",               "figurine1.png",   1+(5*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_KOBOLD_ZOMBIE",     "figurine1.png",   1+(6*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GNOME_ZOMBIE",      "figurine1.png",   1+(7*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_DWARF_ZOMBIE",      "figurine1.png",   1+(8*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ORC_ZOMBIE",        "figurine1.png",   1+(9*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ELF_ZOMBIE",        "figurine1.png",   1+(10*FIG_DX),      1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_KOBOLD_MUMMY",      "figurine1.png",   1+(11*FIG_DX),      1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_HILL_ORC",          "figurine1.png",   1+(0*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_MORDOR_ORC",        "figurine1.png",   1+(1*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_URUK_HAI",          "figurine1.png",   1+(2*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_HOBGOBLIN",         "figurine1.png",   1+(3*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_BUGBEAR",           "figurine1.png",   1+(4*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_KOBOLD",            "figurine1.png",   1+(5*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GNOME_MUMMY",       "figurine1.png",   1+(6*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_DWARF_MUMMY",       "figurine1.png",   1+(7*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ORC_MUMMY",         "figurine1.png",   1+(8*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_HUMAN_MUMMY",       "figurine1.png",   1+(9*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ELF_MUMMY",         "figurine1.png",   1+(10*FIG_DX),      1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_HOMUNCULUS",        "figurine1.png",   1+(11*FIG_DX),      1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LARGE_KOBOLD",      "figurine1.png",   1+(0*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_KOBOLD_SHAMAN",     "figurine1.png",   1+(1*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_KOBOLD_LORD",       "figurine1.png",   1+(2*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ORC_CAPTAIN",       "figurine1.png",   1+(3*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ORC_SHAMAN",        "figurine1.png",   1+(4*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LEPRECHAUN",        "figurine1.png",   1+(5*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_IMP",               "figurine1.png",   1+(6*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LEMURE",            "figurine1.png",   1+(7*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_QUASIT",            "figurine1.png",   1+(8*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_TENGU",             "figurine1.png",   1+(9*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ORACLE",            "figurine1.png",   1+(10*FIG_DX),      1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ALIGNED_PRIEST",    "figurine1.png",   1+(11*FIG_DX),      1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GARTER_SNAKE",      "figurine1.png",   1+(0*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_SNAKE",             "figurine1.png",   1+(1*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_PIT_VIPER",         "figurine1.png",   1+(2*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_PYTHON",            "figurine1.png",   1+(3*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_COBRA",             "figurine1.png",   1+(4*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WUMPUS",            "figurine1.png",   1+(5*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WATCHMAN",          "figurine1.png",   1+(9*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WATCH_CAPTAIN",     "figurine1.png",   1+(10*FIG_DX),      1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_SOLDIER",           "figurine1.png",   1+(11*FIG_DX),      1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_SERGEANT",          "figurine2.png",   1+(0*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LIEUTENANT",        "figurine2.png",   1+(1*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_CAPTAIN",           "figurine2.png",   1+(2*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GRID_BUG",          "figurine2.png",   1+(3*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_XAN",               "figurine2.png",   1+(4*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GARGOYLE",          "figurine2.png",   1+(5*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ARCH_LICH",         "figurine2.png",   1+(6*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GUARDIAN_NAGA",     "figurine2.png",   1+(7*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GUARDIAN_NAGA_HATCHLING","figurine2.png",1+(8*FIG_DX),     1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GOLDEN_NAGA",       "figurine2.png",   1+(9*FIG_DX),       1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GOLDEN_NAGA_HATCHLING","figurine2.png",1+(10*FIG_DX),      1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_RED_NAGA",          "figurine2.png",   1+(11*FIG_DX),      1+(0*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WINGED_GARGOYLE",   "figurine2.png",   1+(0*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_DINGO",             "figurine2.png",   1+(1*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WOLF",              "figurine2.png",   1+(2*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WEREWOLF",          "figurine2.png",   1+(3*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WARG",              "figurine2.png",   1+(4*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WINTER_WOLF_CUB",   "figurine2.png",   1+(5*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_RED_NAGA_HATCHLING","figurine2.png",   1+(6*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_BLACK_NAGA",        "figurine2.png",   1+(7*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_BLACK_NAGA_HATCHLING","figurine2.png", 1+(8*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_RUST_MONSTER",      "figurine2.png",   1+(9*FIG_DX),       1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GREMLIN",           "figurine2.png",   1+(10*FIG_DX),      1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_BABY_PURPLE_WORM",  "figurine2.png",   1+(11*FIG_DX),      1+(1*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WINTER_WOLF",       "figurine2.png",   1+(0*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_HELL_HOUND_PUP",    "figurine2.png",   1+(1*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_HELL_HOUND",        "figurine2.png",   1+(2*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_JAGUAR",            "figurine2.png",   1+(3*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LYNX",              "figurine2.png",   1+(4*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_PANTHER",           "figurine2.png",   1+(5*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_MONKEY",            "figurine2.png",   1+(6*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_APE",               "figurine2.png",   1+(7*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_OWLBEAR",           "figurine2.png",   1+(8*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_YETI",              "figurine2.png",   1+(9*FIG_DX),       1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_CARNIVOROUS_APE",   "figurine2.png",   1+(10*FIG_DX),      1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_STRAW_GOLEM",       "figurine2.png",   1+(11*FIG_DX),      1+(2*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_TIGER",             "figurine2.png",   1+(0*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_MIND_FLAYER",       "figurine2.png",   1+(1*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_MASTER_MIND_FLAYER","figurine2.png",   1+(2*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LICH",              "figurine2.png",   1+(3*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_DEMILICH",          "figurine2.png",   1+(4*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_MASTER_LICH",       "figurine2.png",   1+(5*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_PAPER_GOLEM",       "figurine2.png",   1+(6*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ROPE_GOLEM",        "figurine2.png",   1+(7*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GOLD_GOLEM",        "figurine2.png",   1+(8*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LEATHER_GOLEM",     "figurine2.png",   1+(9*FIG_DX),       1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WOOD_GOLEM",        "figurine2.png",   1+(10*FIG_DX),      1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_FLESH_GOLEM",       "figurine2.png",   1+(11*FIG_DX),      1+(3*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ARCHEOLOGIST",      "figurine2.png",   1+(0*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_TOURIST",           "figurine2.png",   1+(1*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_ROGUE",             "figurine2.png",   1+(2*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_PRIEST",            "figurine2.png",   1+(3*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_COYOTE",            "figurine2.png",   1+(4*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_JACKAL",            "figurine2.png",   1+(5*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_KNIGHT",            "figurine2.png",   1+(6*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GOBLIN",            "figurine2.png",   1+(7*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_QUEEN_BEE",         "figurine2.png",   1+(8*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_FIRE_ELEMENTAL",    "figurine2.png",   1+(9*FIG_DX),       1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GREEN_SLIME",       "figurine2.png",   1+(10*FIG_DX),      1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_SKELETON",          "figurine2.png",   1+(11*FIG_DX),      1+(4*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_VAMPIRE_BAT",       "figurine2.png",   1+(0*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_FOREST_CENTAUR",    "figurine2.png",   1+(1*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_SOLDIER_ANT",       "figurine2.png",   1+(2*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_KILLER_BEE",        "figurine2.png",   1+(3*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_YELLOW_LIGHT",      "figurine2.png",   1+(4*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_FOX",               "figurine2.png",   1+(5*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_FLOATING_EYE",      "figurine2.png",   1+(6*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_QUANTUM_MECHANIC",  "figurine2.png",   1+(7*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_HOUSECAT",          "figurine2.png",   1+(8*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LIZARD",            "figurine2.png",   1+(9*FIG_DX),       1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LITTLE_DOG",        "figurine2.png",   1+(10*FIG_DX),      1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_TROLL",             "figurine2.png",   1+(11*FIG_DX),      1+(5*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_COCKATRICE",        "figurine2.png",   1+(0*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_KRAKEN",            "figurine2.png",   1+(1*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GIANT_EEL",         "figurine2.png",   1+(2*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_PYROLISK",          "figurine2.png",   1+(3*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_YELLOW_MOLD",       "figurine2.png",   1+(4*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_SHARK",             "figurine2.png",   1+(5*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WOOD_NYMPH",        "figurine2.png",   1+(6*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_MOUNTAIN_NYMPH",    "figurine2.png",   1+(7*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WIZARD",            "figurine2.png",   1+(8*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_VALKYRIE",          "figurine2.png",   1+(9*FIG_DX),       1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_RANGER",            "figurine2.png",   1+(10*FIG_DX),      1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WATER_NYMPH",       "figurine2.png",   1+(11*FIG_DX),      1+(6*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_CHICKATRICE",       "figurine2.png",   1+(0*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_SMALL_MIMIC",       "figurine2.png",   1+(1*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_LARGE_MIMIC",       "figurine2.png",   1+(2*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GIANT_MIMIC",       "figurine2.png",   1+(3*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_RED_MOLD",          "figurine2.png",   1+(4*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_VIOLET_FUNGUS",     "figurine2.png",   1+(5*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GHOST",             "figurine2.png",   1+(6*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_WRAITH",            "figurine2.png",   1+(7*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GIANT_BEETLE",      "figurine2.png",   1+(8*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_HUMAN_ZOMBIE",      "figurine2.png",   1+(9*FIG_DX),       1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_CENTIPEDE",         "figurine2.png",   1+(10*FIG_DX),      1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GIANT_RAT",         "figurine2.png",   1+(11*FIG_DX),      1+(7*FIG_DY),       FIG_DX-4,   FIG_DY-4},
{T_FIGURINE,"PM_GIANT",             "figurine4.png",   1+(0*FIG_DX),       1+(0*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_STONE_GIANT",       "figurine4.png",   1+(1*FIG_DX),       1+(0*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_HILL_GIANT",        "figurine4.png",   1+(2*FIG_DX),       1+(0*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_FROST_GIANT",       "figurine4.png",   1+(3*FIG_DX),       1+(0*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_FIRE_GIANT",        "figurine4.png",   1+(4*FIG_DX),       1+(0*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_STORM_GIANT",       "figurine4.png",   1+(5*FIG_DX),       1+(0*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_ETTIN",             "figurine4.png",   1+(0*FIG_DX),       1+(1*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_TITAN",             "figurine4.png",   1+(1*FIG_DX),       1+(1*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_MINOTAUR",          "figurine4.png",   1+(2*FIG_DX),       1+(1*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_GIANT_ZOMBIE",      "figurine4.png",   1+(3*FIG_DX),       1+(1*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_ETTIN_ZOMBIE",      "figurine4.png",   1+(4*FIG_DX),       1+(1*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_GIANT_MUMMY",       "figurine4.png",   1+(5*FIG_DX),       1+(1*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_ETTIN_MUMMY",       "figurine4.png",   1+(0*FIG_DX),       1+(2*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_OGRE",              "figurine4.png",   1+(1*FIG_DX),       1+(2*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_OGRE_LORD",         "figurine4.png",   1+(2*FIG_DX),       1+(2*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_OGRE_KING",         "figurine4.png",   1+(3*FIG_DX),       1+(2*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_DISENCHANTER",      "figurine4.png",   1+(4*FIG_DX),       1+(2*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_SASQUATCH",         "figurine4.png",   1+(5*FIG_DX),       1+(2*FIG_LARGE_DY), FIG_DX-4,   FIG_LARGE_DY-4},
{T_FIGURINE,"PM_SUCCUBUS",          "figurine5.png",   1+(0*FIG_LARGE_DX), 1+(0*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_INCUBUS",           "figurine5.png",   1+(1*FIG_LARGE_DX), 1+(0*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_BABY_YELLOW_DRAGON","figurine5.png",   1+(2*FIG_LARGE_DX), 1+(0*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_BABY_WHITE_DRAGON", "figurine5.png",   1+(3*FIG_LARGE_DX), 1+(0*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_BABY_SILVER_DRAGON","figurine5.png",   1+(4*FIG_LARGE_DX), 1+(0*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_BABY_RED_DRAGON",   "figurine5.png",   1+(0*FIG_LARGE_DX), 1+(1*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_BABY_BLACK_DRAGON", "figurine5.png",   1+(1*FIG_LARGE_DX), 1+(1*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_BABY_GRAY_DRAGON",  "figurine5.png",   1+(2*FIG_LARGE_DX), 1+(1*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_BABY_GREEN_DRAGON", "figurine5.png",   1+(3*FIG_LARGE_DX), 1+(1*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_BABY_ORANGE_DRAGON","figurine5.png",   1+(4*FIG_LARGE_DX), 1+(1*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_BABY_BLUE_DRAGON",  "figurine5.png",   1+(0*FIG_LARGE_DX), 1+(2*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_PURPLE_WORM",       "figurine5.png",   1+(1*FIG_LARGE_DX), 1+(2*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_LONG_WORM_TAIL",    "figurine5.png",   1+(2*FIG_LARGE_DX), 1+(2*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_LONG_WORM",         "figurine5.png",   1+(3*FIG_LARGE_DX), 1+(2*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_BABY_LONG_WORM",    "figurine5.png",   1+(4*FIG_LARGE_DX), 1+(2*FIG_LARGE_DY), FIG_LARGE_DX-4,FIG_LARGE_DY-4},
{T_FIGURINE,"PM_SILVER_DRAGON",     "figurine6.png",   1+(0*FIG_HUGE_DX),  1+(0*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_BLACK_DRAGON",      "figurine6.png",   1+(1*FIG_HUGE_DX),  1+(0*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_WHITE_DRAGON",      "figurine6.png",   1+(2*FIG_HUGE_DX),  1+(0*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_BALUCHITHERIUM",    "figurine6.png",   1+(0*FIG_HUGE_DX),  1+(1*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_MASTODON",          "figurine6.png",   1+(1*FIG_HUGE_DX),  1+(1*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_WIZARD_OF_YENDOR",  "figurine6.png",   1+(2*FIG_HUGE_DX),  1+(1*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_RED_DRAGON",        "figurine6.png",   1+(3*FIG_HUGE_DX),  1+(0*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_GREEN_DRAGON",      "figurine6.png",   1+(4*FIG_HUGE_DX),  1+(0*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_ORANGE_DRAGON",     "figurine6.png",   1+(5*FIG_HUGE_DX),  1+(0*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_YELLOW_DRAGON",     "figurine6.png",   1+(3*FIG_HUGE_DX),  1+(1*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_BLUE_DRAGON",       "figurine6.png",   1+(4*FIG_HUGE_DX),  1+(1*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},
{T_FIGURINE,"PM_GRAY_DRAGON",       "figurine6.png",   1+(5*FIG_HUGE_DX),  1+(1*FIG_HUGE_DY),  FIG_HUGE_DX-4, FIG_HUGE_DY-4},

/* statues: NAME                   FILENAME          top left x coord     top left y coord    width       height*/
/* The name field is the name of a define in pm.h */
{T_STATUE, "PM_KNIGHT",            "statue6.png",   1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GOBLIN",            "statue6.png",   1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_QUEEN_BEE",         "statue6.png",   1+(2*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_FIRE_ELEMENTAL",    "statue6.png",   1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GREEN_SLIME",       "statue6.png",   1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_SKELETON",          "statue6.png",   1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_FLOATING_EYE",      "statue6.png",   1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_QUANTUM_MECHANIC",  "statue6.png",   1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_HOUSECAT",          "statue6.png",   1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LIZARD",            "statue6.png",   1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LITTLE_DOG",        "statue6.png",   1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_TROLL",             "statue6.png",   1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WOOD_NYMPH",        "statue6.png",   1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_MOUNTAIN_NYMPH",    "statue6.png",   1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WIZARD",            "statue6.png",   1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_VALKYRIE",          "statue6.png",   1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_RANGER",            "statue6.png",   1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WATER_NYMPH",       "statue6.png",   1+(5*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GHOST",             "statue6.png",   1+(0*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WRAITH",            "statue6.png",   1+(1*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GIANT_BEETLE",      "statue6.png",   1+(2*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_HUMAN_ZOMBIE",      "statue6.png",   1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_CENTIPEDE",         "statue6.png",   1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GIANT_RAT",         "statue6.png",   1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ARCHEOLOGIST",      "statue7.png",   1+(0*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_TOURIST",           "statue7.png",   1+(1*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ROGUE",             "statue7.png",   1+(2*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_PRIEST",            "statue7.png",   1+(3*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_COYOTE",            "statue7.png",   1+(4*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_JACKAL",            "statue7.png",   1+(5*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_VAMPIRE_BAT",       "statue7.png",   1+(0*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_FOREST_CENTAUR",    "statue7.png",   1+(1*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_SOLDIER_ANT",       "statue7.png",   1+(2*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_KILLER_BEE",        "statue7.png",   1+(3*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_YELLOW_LIGHT",      "statue7.png",   1+(4*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_FOX",               "statue7.png",   1+(5*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_COCKATRICE",        "statue7.png",   1+(0*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_KRAKEN",            "statue7.png",   1+(1*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GIANT_EEL",         "statue7.png",   1+(2*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_PYROLISK",          "statue7.png",   1+(3*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_YELLOW_MOLD",       "statue7.png",   1+(4*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_SHARK",             "statue7.png",   1+(5*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_CHICKATRICE",       "statue7.png",   1+(0*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_SMALL_MIMIC",       "statue7.png",   1+(1*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LARGE_MIMIC",       "statue7.png",   1+(2*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GIANT_MIMIC",       "statue7.png",   1+(3*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_RED_MOLD",          "statue7.png",   1+(4*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_VIOLET_FUNGUS",     "statue7.png",   1+(5*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ACID_BLOB",         "statue8.png",   1+(0*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_QUIVERING_BLOB",    "statue8.png",   1+(1*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GELATINOUS_CUBE",   "statue8.png",   1+(2*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_BROWN_MOLD",        "statue8.png",   1+(3*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LICHEN",            "statue8.png",   1+(4*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GREEN_MOLD",        "statue8.png",   1+(5*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_SHRIEKER",          "statue8.png",   1+(0*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_STALKER",           "statue8.png",   1+(1*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_AIR_ELEMENTAL",     "statue8.png",   1+(2*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_EARTH_ELEMENTAL",   "statue8.png",   1+(3*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WATER_ELEMENTAL",   "statue8.png",   1+(4*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_BLACK_LIGHT",       "statue8.png",   1+(5*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ROCK_PIERCER",      "statue8.png",   1+(0*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_IRON_PIERCER",      "statue8.png",   1+(1*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GLASS_PIERCER",     "statue8.png",   1+(2*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_DOG",               "statue8.png",   1+(3*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LARGE_DOG",         "statue8.png",   1+(4*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WATER_DEMON",       "statue8.png",   1+(5*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GECKO",             "statue8.png",   1+(0*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_NEWT",              "statue8.png",   1+(1*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GAS_SPORE",         "statue8.png",   1+(3*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_KITTEN",            "statue8.png",   1+(4*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LARGE_CAT",         "statue8.png",   1+(5*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_FREEZING_SPHERE",   "statue9.png",   1+(0*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_FLAMING_SPHERE",    "statue9.png",   1+(1*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_SHOCKING_SPHERE",   "statue9.png",   1+(2*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_PONY",              "statue9.png",   1+(3*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_HORSE",             "statue9.png",   1+(4*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WARHORSE",          "statue9.png",   1+(5*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ROTHE",             "statue9.png",   1+(0*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GNOME",             "statue9.png",   1+(1*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GREEN_ELF",         "statue9.png",   1+(2*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_DWARF",             "statue9.png",   1+(3*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WATER_MOCCASIN",    "statue9.png",   1+(4*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_MANES",             "statue9.png",   1+(5*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_HOBBIT",            "statue9.png",   1+(0*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_BLACK_UNICORN",     "statue9.png",   1+(1*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GRAY_UNICORN",      "statue9.png",   1+(2*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WHITE_UNICORN",     "statue9.png",   1+(3*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GIANT_ANT",         "statue9.png",   1+(4*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_FIRE_ANT",          "statue9.png",   1+(5*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ELVENKING",         "statue9.png",   1+(0*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ELF_LORD",          "statue9.png",   1+(1*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GREY_ELF",          "statue9.png",   1+(2*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WOODLAND_ELF",      "statue9.png",   1+(3*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ELF",               "statue9.png",   1+(4*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ROCK_MOLE",         "statue9.png",   1+(5*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_DWARF_LORD",        "statue0.png",   1+(0*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_DWARF_KING",        "statue0.png",   1+(1*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GNOME_LORD",        "statue0.png",   1+(2*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GNOME_KING",        "statue0.png",   1+(3*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GNOMISH_WIZARD",    "statue0.png",   1+(4*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ORC",               "statue0.png",   1+(5*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_HILL_ORC",          "statue0.png",   1+(0*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_MORDOR_ORC",        "statue0.png",   1+(1*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_URUK_HAI",          "statue0.png",   1+(2*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_HOBGOBLIN",         "statue0.png",   1+(3*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_BUGBEAR",           "statue0.png",   1+(4*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_KOBOLD",            "statue0.png",   1+(5*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LARGE_KOBOLD",      "statue0.png",   1+(0*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_KOBOLD_SHAMAN",     "statue0.png",   1+(1*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_KOBOLD_LORD",       "statue0.png",   1+(2*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ORC_CAPTAIN",       "statue0.png",   1+(3*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ORC_SHAMAN",        "statue0.png",   1+(4*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LEPRECHAUN",        "statue0.png",   1+(5*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GARTER_SNAKE",      "statue0.png",   1+(0*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_SNAKE",             "statue0.png",   1+(1*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_PIT_VIPER",         "statue0.png",   1+(2*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_PYTHON",            "statue0.png",   1+(3*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_COBRA",             "statue0.png",   1+(4*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WUMPUS",            "statue0.png",   1+(5*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_KOBOLD_ZOMBIE",     "statue1.png",   1+(0*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GNOME_ZOMBIE",      "statue1.png",   1+(1*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_DWARF_ZOMBIE",      "statue1.png",   1+(2*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ORC_ZOMBIE",        "statue1.png",   1+(3*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ELF_ZOMBIE",        "statue1.png",   1+(4*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_KOBOLD_MUMMY",      "statue1.png",   1+(5*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GNOME_MUMMY",       "statue1.png",   1+(0*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_DWARF_MUMMY",       "statue1.png",   1+(1*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ORC_MUMMY",         "statue1.png",   1+(2*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_HUMAN_MUMMY",       "statue1.png",   1+(3*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ELF_MUMMY",         "statue1.png",   1+(4*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_HOMUNCULUS",        "statue1.png",   1+(5*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_IMP",               "statue1.png",   1+(0*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LEMURE",            "statue1.png",   1+(1*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_QUASIT",            "statue1.png",   1+(2*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_TENGU",             "statue1.png",   1+(3*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ORACLE",            "statue1.png",   1+(4*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ALIGNED_PRIEST",    "statue1.png",   1+(5*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WATCHMAN",          "statue1.png",   1+(3*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WATCH_CAPTAIN",     "statue1.png",   1+(4*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_SOLDIER",           "statue1.png",   1+(5*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_SERGEANT",          "statue2.png",   1+(0*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LIEUTENANT",        "statue2.png",   1+(1*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_CAPTAIN",           "statue2.png",   1+(2*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GRID_BUG",          "statue2.png",   1+(3*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_XAN",               "statue2.png",   1+(4*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GARGOYLE",          "statue2.png",   1+(5*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WINGED_GARGOYLE",   "statue2.png",   1+(0*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_DINGO",             "statue2.png",   1+(1*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WOLF",              "statue2.png",   1+(2*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WEREWOLF",          "statue2.png",   1+(3*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WARG",              "statue2.png",   1+(4*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WINTER_WOLF_CUB",   "statue2.png",   1+(5*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WINTER_WOLF",       "statue2.png",   1+(0*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_HELL_HOUND_PUP",    "statue2.png",   1+(1*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_HELL_HOUND",        "statue2.png",   1+(2*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_JAGUAR",            "statue2.png",   1+(3*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LYNX",              "statue2.png",   1+(4*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_PANTHER",           "statue2.png",   1+(5*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_TIGER",             "statue2.png",   1+(0*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_MIND_FLAYER",       "statue2.png",   1+(1*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_MASTER_MIND_FLAYER","statue2.png",   1+(2*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LICH",              "statue2.png",   1+(3*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_DEMILICH",          "statue2.png",   1+(4*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_MASTER_LICH",       "statue2.png",   1+(5*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ARCH_LICH",         "statue3.png",   1+(0*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GUARDIAN_NAGA",     "statue3.png",   1+(1*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GUARDIAN_NAGA_HATCHLING","statue3.png",1+(2*MON_DX),       1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GOLDEN_NAGA",       "statue3.png",   1+(3*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GOLDEN_NAGA_HATCHLING","statue3.png",1+(4*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_RED_NAGA",          "statue3.png",   1+(5*MON_DX),         1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_RED_NAGA_HATCHLING","statue3.png",   1+(0*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_BLACK_NAGA",        "statue3.png",   1+(1*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_BLACK_NAGA_HATCHLING","statue3.png", 1+(2*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_RUST_MONSTER",      "statue3.png",   1+(3*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GREMLIN",           "statue3.png",   1+(4*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_BABY_PURPLE_WORM",  "statue3.png",   1+(5*MON_DX),         1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_MONKEY",            "statue3.png",   1+(0*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_APE",               "statue3.png",   1+(1*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_OWLBEAR",           "statue3.png",   1+(2*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_YETI",              "statue3.png",   1+(3*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_CARNIVOROUS_APE",   "statue3.png",   1+(4*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_STRAW_GOLEM",       "statue3.png",   1+(5*MON_DX),         1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_PAPER_GOLEM",       "statue3.png",   1+(0*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_ROPE_GOLEM",        "statue3.png",   1+(1*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GOLD_GOLEM",        "statue3.png",   1+(2*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_LEATHER_GOLEM",     "statue3.png",   1+(3*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_WOOD_GOLEM",        "statue3.png",   1+(4*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_FLESH_GOLEM",       "statue3.png",   1+(5*MON_DX),         1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_STATUE, "PM_GIANT",             "statue4.png",   1+(0*MON_DX),         1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_STONE_GIANT",       "statue4.png",   1+(1*MON_DX),         1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_HILL_GIANT",        "statue4.png",   1+(2*MON_DX),         1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_FROST_GIANT",       "statue4.png",   1+(3*MON_DX),         1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_FIRE_GIANT",        "statue4.png",   1+(4*MON_DX),         1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_STORM_GIANT",       "statue4.png",   1+(5*MON_DX),         1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_ETTIN",             "statue4.png",   1+(0*MON_DX),         1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_TITAN",             "statue4.png",   1+(1*MON_DX),         1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_MINOTAUR",          "statue4.png",   1+(2*MON_DX),         1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_GIANT_ZOMBIE",      "statue4.png",   1+(3*MON_DX),         1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_ETTIN_ZOMBIE",      "statue4.png",   1+(4*MON_DX),         1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_GIANT_MUMMY",       "statue4.png",   1+(5*MON_DX),         1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_ETTIN_MUMMY",       "statue4.png",   1+(0*MON_DX),         1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_OGRE",              "statue4.png",   1+(1*MON_DX),         1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_OGRE_LORD",         "statue4.png",   1+(2*MON_DX),         1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_OGRE_KING",         "statue4.png",   1+(3*MON_DX),         1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_DISENCHANTER",      "statue4.png",   1+(4*MON_DX),         1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_SASQUATCH",         "statue4.png",   1+(5*MON_DX),         1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_STATUE, "PM_SUCCUBUS",          "statue5.png",   1+(0*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_INCUBUS",           "statue5.png",   1+(1*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_BABY_YELLOW_DRAGON","statue5.png",   1+(2*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_BABY_WHITE_DRAGON", "statue5.png",   1+(3*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_BABY_SILVER_DRAGON","statue5.png",   1+(4*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_BABY_RED_DRAGON",   "statue5.png",   1+(0*MON_LARGE_DX),  1+(1*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_BABY_BLACK_DRAGON", "statue5.png",   1+(1*MON_LARGE_DX),  1+(1*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_BABY_GRAY_DRAGON",  "statue5.png",   1+(2*MON_LARGE_DX),  1+(1*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_BABY_GREEN_DRAGON", "statue5.png",   1+(3*MON_LARGE_DX),  1+(1*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_BABY_ORANGE_DRAGON","statue5.png",   1+(4*MON_LARGE_DX),  1+(1*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_BABY_BLUE_DRAGON",  "statue5.png",   1+(0*MON_LARGE_DX),  1+(2*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_PURPLE_WORM",       "statue5.png",   1+(1*MON_LARGE_DX),  1+(2*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_LONG_WORM_TAIL",    "statue5.png",   1+(2*MON_LARGE_DX),  1+(2*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_LONG_WORM",         "statue5.png",   1+(3*MON_LARGE_DX),  1+(2*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_BABY_LONG_WORM",    "statue5.png",   1+(4*MON_LARGE_DX),  1+(2*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_STATUE, "PM_SILVER_DRAGON",     "statue14.png",  1+(0*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_BLACK_DRAGON",      "statue14.png",  1+(1*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_WHITE_DRAGON",      "statue14.png",  1+(2*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_BALUCHITHERIUM",    "statue14.png",  1+(0*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_MASTODON",          "statue14.png",  1+(1*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_WIZARD_OF_YENDOR",  "statue14.png",  1+(2*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_RED_DRAGON",        "statue15.png",  1+(0*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_GREEN_DRAGON",      "statue15.png",  1+(1*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_ORANGE_DRAGON",     "statue15.png",  1+(2*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_YELLOW_DRAGON",     "statue15.png",  1+(0*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_BLUE_DRAGON",       "statue15.png",  1+(1*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},
{T_STATUE, "PM_GRAY_DRAGON",       "statue15.png",  1+(2*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4, MON_HUGE_DY-4},

/* statues: NAME                   FILENAME          top left x coord     top left y coord    width       height*/
/* The name field is the name of a define in pm.h */
{T_MONSTER,"PM_KNIGHT",            "jtp_mon6.png",   1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GOBLIN",            "jtp_mon6.png",   1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_QUEEN_BEE",         "jtp_mon6.png",   1+(2*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_FIRE_ELEMENTAL",    "jtp_mon6.png",   1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GREEN_SLIME",       "jtp_mon6.png",   1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SKELETON",          "jtp_mon6.png",   1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_FLOATING_EYE",      "jtp_mon6.png",   1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_QUANTUM_MECHANIC",  "jtp_mon6.png",   1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HOUSECAT",          "jtp_mon6.png",   1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LIZARD",            "jtp_mon6.png",   1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LITTLE_DOG",        "jtp_mon6.png",   1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_TROLL",             "jtp_mon6.png",   1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WOOD_NYMPH",        "jtp_mon6.png",   1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_MOUNTAIN_NYMPH",    "jtp_mon6.png",   1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WIZARD",            "jtp_mon6.png",   1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_VALKYRIE",          "jtp_mon6.png",   1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_RANGER",            "jtp_mon6.png",   1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WATER_NYMPH",       "jtp_mon6.png",   1+(5*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GHOST",             "jtp_mon6.png",   1+(0*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WRAITH",            "jtp_mon6.png",   1+(1*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GIANT_BEETLE",      "jtp_mon6.png",   1+(2*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HUMAN_ZOMBIE",      "jtp_mon6.png",   1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_CENTIPEDE",         "jtp_mon6.png",   1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GIANT_RAT",         "jtp_mon6.png",   1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ARCHEOLOGIST",      "jtp_mon7.png",   1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_TOURIST",           "jtp_mon7.png",   1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ROGUE",             "jtp_mon7.png",   1+(2*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_PRIEST",            "jtp_mon7.png",   1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_COYOTE",            "jtp_mon7.png",   1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_JACKAL",            "jtp_mon7.png",   1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_VAMPIRE_BAT",       "jtp_mon7.png",   1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_FOREST_CENTAUR",    "jtp_mon7.png",   1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SOLDIER_ANT",       "jtp_mon7.png",   1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KILLER_BEE",        "jtp_mon7.png",   1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_YELLOW_LIGHT",      "jtp_mon7.png",   1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_FOX",               "jtp_mon7.png",   1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_COCKATRICE",        "jtp_mon7.png",   1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KRAKEN",            "jtp_mon7.png",   1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GIANT_EEL",         "jtp_mon7.png",   1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_PYROLISK",          "jtp_mon7.png",   1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_YELLOW_MOLD",       "jtp_mon7.png",   1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SHARK",             "jtp_mon7.png",   1+(5*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_CHICKATRICE",       "jtp_mon7.png",   1+(0*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SMALL_MIMIC",       "jtp_mon7.png",   1+(1*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LARGE_MIMIC",       "jtp_mon7.png",   1+(2*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GIANT_MIMIC",       "jtp_mon7.png",   1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_RED_MOLD",          "jtp_mon7.png",   1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_VIOLET_FUNGUS",     "jtp_mon7.png",   1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ACID_BLOB",         "jtp_mon8.png",   1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_QUIVERING_BLOB",    "jtp_mon8.png",   1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GELATINOUS_CUBE",   "jtp_mon8.png",   1+(2*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_BROWN_MOLD",        "jtp_mon8.png",   1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LICHEN",            "jtp_mon8.png",   1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GREEN_MOLD",        "jtp_mon8.png",   1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SHRIEKER",          "jtp_mon8.png",   1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_STALKER",           "jtp_mon8.png",   1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_AIR_ELEMENTAL",     "jtp_mon8.png",   1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_EARTH_ELEMENTAL",   "jtp_mon8.png",   1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WATER_ELEMENTAL",   "jtp_mon8.png",   1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_BLACK_LIGHT",       "jtp_mon8.png",   1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ROCK_PIERCER",      "jtp_mon8.png",   1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_IRON_PIERCER",      "jtp_mon8.png",   1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GLASS_PIERCER",     "jtp_mon8.png",   1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_DOG",               "jtp_mon8.png",   1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LARGE_DOG",         "jtp_mon8.png",   1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WATER_DEMON",       "jtp_mon8.png",   1+(5*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GECKO",             "jtp_mon8.png",   1+(0*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_NEWT",              "jtp_mon8.png",   1+(1*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SHOPKEEPER",        "jtp_mon8.png",   1+(2*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GAS_SPORE",         "jtp_mon8.png",   1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KITTEN",            "jtp_mon8.png",   1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LARGE_CAT",         "jtp_mon8.png",   1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_FREEZING_SPHERE",   "jtp_mon9.png",   1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_FLAMING_SPHERE",    "jtp_mon9.png",   1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SHOCKING_SPHERE",   "jtp_mon9.png",   1+(2*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_PONY",              "jtp_mon9.png",   1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HORSE",             "jtp_mon9.png",   1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WARHORSE",          "jtp_mon9.png",   1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ROTHE",             "jtp_mon9.png",   1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GNOME",             "jtp_mon9.png",   1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GREEN_ELF",         "jtp_mon9.png",   1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_DWARF",             "jtp_mon9.png",   1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WATER_MOCCASIN",    "jtp_mon9.png",   1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_MANES",             "jtp_mon9.png",   1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HOBBIT",            "jtp_mon9.png",   1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_BLACK_UNICORN",     "jtp_mon9.png",   1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GRAY_UNICORN",      "jtp_mon9.png",   1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WHITE_UNICORN",     "jtp_mon9.png",   1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GIANT_ANT",         "jtp_mon9.png",   1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_FIRE_ANT",          "jtp_mon9.png",   1+(5*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ELVENKING",         "jtp_mon9.png",   1+(0*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ELF_LORD",          "jtp_mon9.png",   1+(1*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GREY_ELF",          "jtp_mon9.png",   1+(2*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WOODLAND_ELF",      "jtp_mon9.png",   1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ELF",               "jtp_mon9.png",   1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ROCK_MOLE",         "jtp_mon9.png",   1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_DWARF_LORD",        "jtp_mon0.png",   1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_DWARF_KING",        "jtp_mon0.png",   1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GNOME_LORD",        "jtp_mon0.png",   1+(2*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GNOME_KING",        "jtp_mon0.png",   1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GNOMISH_WIZARD",    "jtp_mon0.png",   1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ORC",               "jtp_mon0.png",   1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HILL_ORC",          "jtp_mon0.png",   1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_MORDOR_ORC",        "jtp_mon0.png",   1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_URUK_HAI",          "jtp_mon0.png",   1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HOBGOBLIN",         "jtp_mon0.png",   1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_BUGBEAR",           "jtp_mon0.png",   1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KOBOLD",            "jtp_mon0.png",   1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LARGE_KOBOLD",      "jtp_mon0.png",   1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KOBOLD_SHAMAN",     "jtp_mon0.png",   1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KOBOLD_LORD",       "jtp_mon0.png",   1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ORC_CAPTAIN",       "jtp_mon0.png",   1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ORC_SHAMAN",        "jtp_mon0.png",   1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LEPRECHAUN",        "jtp_mon0.png",   1+(5*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GARTER_SNAKE",      "jtp_mon0.png",   1+(0*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SNAKE",             "jtp_mon0.png",   1+(1*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_PIT_VIPER",         "jtp_mon0.png",   1+(2*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_PYTHON",            "jtp_mon0.png",   1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_COBRA",             "jtp_mon0.png",   1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WUMPUS",            "jtp_mon0.png",   1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KOBOLD_ZOMBIE",     "jtp_mon1.png",   1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GNOME_ZOMBIE",      "jtp_mon1.png",   1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_DWARF_ZOMBIE",      "jtp_mon1.png",   1+(2*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ORC_ZOMBIE",        "jtp_mon1.png",   1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ELF_ZOMBIE",        "jtp_mon1.png",   1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KOBOLD_MUMMY",      "jtp_mon1.png",   1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GNOME_MUMMY",       "jtp_mon1.png",   1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_DWARF_MUMMY",       "jtp_mon1.png",   1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ORC_MUMMY",         "jtp_mon1.png",   1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HUMAN_MUMMY",       "jtp_mon1.png",   1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ELF_MUMMY",         "jtp_mon1.png",   1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HOMUNCULUS",        "jtp_mon1.png",   1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_IMP",               "jtp_mon1.png",   1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LEMURE",            "jtp_mon1.png",   1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_QUASIT",            "jtp_mon1.png",   1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_TENGU",             "jtp_mon1.png",   1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ORACLE",            "jtp_mon1.png",   1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WATCHMAN",          "jtp_mon1.png",   1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WATCH_CAPTAIN",     "jtp_mon1.png",   1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SOLDIER",           "jtp_mon1.png",   1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SERGEANT",          "jtp_mon2.png",   1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LIEUTENANT",        "jtp_mon2.png",   1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_CAPTAIN",           "jtp_mon2.png",   1+(2*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GRID_BUG",          "jtp_mon2.png",   1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_XAN",               "jtp_mon2.png",   1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GARGOYLE",          "jtp_mon2.png",   1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WINGED_GARGOYLE",   "jtp_mon2.png",   1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_DINGO",             "jtp_mon2.png",   1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WOLF",              "jtp_mon2.png",   1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WEREWOLF",          "jtp_mon2.png",   1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WARG",              "jtp_mon2.png",   1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WINTER_WOLF_CUB",   "jtp_mon2.png",   1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WINTER_WOLF",       "jtp_mon2.png",   1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HELL_HOUND_PUP",    "jtp_mon2.png",   1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HELL_HOUND",        "jtp_mon2.png",   1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_JAGUAR",            "jtp_mon2.png",   1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LYNX",              "jtp_mon2.png",   1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_PANTHER",           "jtp_mon2.png",   1+(5*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_TIGER",             "jtp_mon2.png",   1+(0*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_MIND_FLAYER",       "jtp_mon2.png",   1+(1*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_MASTER_MIND_FLAYER","jtp_mon2.png",   1+(2*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LICH",              "jtp_mon2.png",   1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_DEMILICH",          "jtp_mon2.png",   1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_MASTER_LICH",       "jtp_mon2.png",   1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ARCH_LICH",         "jtp_mon3.png",   1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GUARDIAN_NAGA",     "jtp_mon3.png",   1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GUARDIAN_NAGA_HATCHLING","jtp_mon3.png",1+(2*MON_DX),      1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GOLDEN_NAGA",       "jtp_mon3.png",   1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GOLDEN_NAGA_HATCHLING","jtp_mon3.png",1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_RED_NAGA",          "jtp_mon3.png",   1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_RED_NAGA_HATCHLING","jtp_mon3.png",   1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_BLACK_NAGA",        "jtp_mon3.png",   1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_BLACK_NAGA_HATCHLING","jtp_mon3.png", 1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_RUST_MONSTER",      "jtp_mon3.png",   1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GREMLIN",           "jtp_mon3.png",   1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_BABY_PURPLE_WORM",  "jtp_mon3.png",   1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_MONKEY",            "jtp_mon3.png",   1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_APE",               "jtp_mon3.png",   1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_OWLBEAR",           "jtp_mon3.png",   1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_YETI",              "jtp_mon3.png",   1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_CARNIVOROUS_APE",   "jtp_mon3.png",   1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_STRAW_GOLEM",       "jtp_mon3.png",   1+(5*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_PAPER_GOLEM",       "jtp_mon3.png",   1+(0*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_ROPE_GOLEM",        "jtp_mon3.png",   1+(1*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GOLD_GOLEM",        "jtp_mon3.png",   1+(2*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_LEATHER_GOLEM",     "jtp_mon3.png",   1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WOOD_GOLEM",        "jtp_mon3.png",   1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_FLESH_GOLEM",       "jtp_mon3.png",   1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SCORPION",          "jtp_mon10.png",  1+(0*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_CAVE_SPIDER",       "jtp_mon10.png",  1+(1*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GIANT_SPIDER",      "jtp_mon10.png",  1+(2*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_BAT",               "jtp_mon10.png",  1+(3*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GIANT_BAT",         "jtp_mon10.png",  1+(4*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_RAVEN",             "jtp_mon10.png",  1+(5*MON_DX),        1+(0*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_DJINNI",            "jtp_mon10.png",  1+(0*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_MOUNTAIN_CENTAUR",  "jtp_mon10.png",  1+(1*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_PLAINS_CENTAUR",    "jtp_mon10.png",  1+(2*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_COUATL",            "jtp_mon10.png",  1+(3*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_DOPPELGANGER",      "jtp_mon10.png",  1+(4*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_CROESUS",           "jtp_mon10.png",  1+(5*MON_DX),        1+(1*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KEYSTONE_KOP",      "jtp_mon10.png",  1+(0*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KOP_SERGEANT",      "jtp_mon10.png",  1+(1*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KOP_LIEUTENANT",    "jtp_mon10.png",  1+(2*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_KOP_KAPTAIN",       "jtp_mon10.png",  1+(3*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_VAMPIRE",           "jtp_mon10.png",  1+(4*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_SEWER_RAT",         "jtp_mon10.png",  1+(5*MON_DX),        1+(2*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_RABID_RAT",         "jtp_mon10.png",  1+(0*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WERERAT",           "jtp_mon10.png",  1+(1*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_WOODCHUCK",         "jtp_mon10.png",  1+(2*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_MEDUSA",            "jtp_mon10.png",  1+(3*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_HIGH_PRIEST",       "jtp_mon10.png",  1+(4*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_NURSE",             "jtp_mon10.png",  1+(5*MON_DX),        1+(3*MON_DY),       MON_DX-4,   MON_DY-4},
{T_MONSTER,"PM_GIANT",             "jtp_mon4.png",   1+(0*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_STONE_GIANT",       "jtp_mon4.png",   1+(1*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_HILL_GIANT",        "jtp_mon4.png",   1+(2*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_FROST_GIANT",       "jtp_mon4.png",   1+(3*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_FIRE_GIANT",        "jtp_mon4.png",   1+(4*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_STORM_GIANT",       "jtp_mon4.png",   1+(5*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_ETTIN",             "jtp_mon4.png",   1+(0*MON_DX),        1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_TITAN",             "jtp_mon4.png",   1+(1*MON_DX),        1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_MINOTAUR",          "jtp_mon4.png",   1+(2*MON_DX),        1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_GIANT_ZOMBIE",      "jtp_mon4.png",   1+(3*MON_DX),        1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_ETTIN_ZOMBIE",      "jtp_mon4.png",   1+(4*MON_DX),        1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_GIANT_MUMMY",       "jtp_mon4.png",   1+(5*MON_DX),        1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_ETTIN_MUMMY",       "jtp_mon4.png",   1+(0*MON_DX),        1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_OGRE",              "jtp_mon4.png",   1+(1*MON_DX),        1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_OGRE_LORD",         "jtp_mon4.png",   1+(2*MON_DX),        1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_OGRE_KING",         "jtp_mon4.png",   1+(3*MON_DX),        1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_DISENCHANTER",      "jtp_mon4.png",   1+(4*MON_DX),        1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_SASQUATCH",         "jtp_mon4.png",   1+(5*MON_DX),        1+(2*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_SUCCUBUS",          "jtp_mon5.png",   1+(0*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_INCUBUS",           "jtp_mon5.png",   1+(1*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_BABY_YELLOW_DRAGON","jtp_mon5.png",   1+(2*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_BABY_WHITE_DRAGON", "jtp_mon5.png",   1+(3*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_BABY_SILVER_DRAGON","jtp_mon5.png",   1+(4*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_BABY_RED_DRAGON",   "jtp_mon5.png",   1+(0*MON_LARGE_DX),  1+(1*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_BABY_BLACK_DRAGON", "jtp_mon5.png",   1+(1*MON_LARGE_DX),  1+(1*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_BABY_GRAY_DRAGON",  "jtp_mon5.png",   1+(2*MON_LARGE_DX),  1+(1*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_BABY_GREEN_DRAGON", "jtp_mon5.png",   1+(3*MON_LARGE_DX),  1+(1*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_BABY_ORANGE_DRAGON","jtp_mon5.png",   1+(4*MON_LARGE_DX),  1+(1*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_BABY_BLUE_DRAGON",  "jtp_mon5.png",   1+(0*MON_LARGE_DX),  1+(2*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_PURPLE_WORM",       "jtp_mon5.png",   1+(1*MON_LARGE_DX),  1+(2*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_LONG_WORM_TAIL",    "jtp_mon5.png",   1+(2*MON_LARGE_DX),  1+(2*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_LONG_WORM",         "jtp_mon5.png",   1+(3*MON_LARGE_DX),  1+(2*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_BABY_LONG_WORM",    "jtp_mon5.png",   1+(4*MON_LARGE_DX),  1+(2*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_MUMAK",             "jtp_mon16.png",  1+(0*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_CLAY_GOLEM",        "jtp_mon16.png",  1+(1*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_STONE_GOLEM",       "jtp_mon16.png",  1+(2*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_GLASS_GOLEM",       "jtp_mon16.png",  1+(3*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_IRON_GOLEM",        "jtp_mon16.png",  1+(4*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_LURKER_ABOVE",      "jtp_mon16.png",  1+(5*MON_DX),        1+(0*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_TRAPPER",           "jtp_mon16.png",  1+(0*MON_DX),        1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_ROCK_TROLL",        "jtp_mon16.png",  1+(1*MON_DX),        1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_WATER_TROLL",       "jtp_mon16.png",  1+(2*MON_DX),        1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_OLOG_HAI",          "jtp_mon16.png",  1+(3*MON_DX),        1+(1*MON_LARGE_DY), MON_DX-4,   MON_LARGE_DY-4},
{T_MONSTER,"PM_ZRUTY",             "jtp_mon18.png",  1+(0*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_LEOCROTTA",         "jtp_mon18.png",  1+(1*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_TITANOTHERE",       "jtp_mon18.png",  1+(2*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_XORN",              "jtp_mon18.png",  1+(3*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_JABBERWOCK",        "jtp_mon18.png",  1+(4*MON_LARGE_DX),  1+(0*MON_LARGE_DY), MON_LARGE_DX-4,MON_LARGE_DY-4},
{T_MONSTER,"PM_ANGEL",             "jtp_mon13.png",  1+(0*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_KI_RIN",            "jtp_mon13.png",  1+(1*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_ARCHON",            "jtp_mon13.png",  1+(2*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_SILVER_DRAGON",     "jtp_mon14.png",  1+(0*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_BLACK_DRAGON",      "jtp_mon14.png",  1+(1*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_WHITE_DRAGON",      "jtp_mon14.png",  1+(2*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_BALUCHITHERIUM",    "jtp_mon14.png",  1+(0*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_MASTODON",          "jtp_mon14.png",  1+(1*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_WIZARD_OF_YENDOR",  "jtp_mon14.png",  1+(2*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_RED_DRAGON",        "jtp_mon15.png",  1+(0*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_GREEN_DRAGON",      "jtp_mon15.png",  1+(1*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_ORANGE_DRAGON",     "jtp_mon15.png",  1+(2*MON_HUGE_DX),   1+(0*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_YELLOW_DRAGON",     "jtp_mon15.png",  1+(0*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_BLUE_DRAGON",       "jtp_mon15.png",  1+(1*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},
{T_MONSTER,"PM_GRAY_DRAGON",       "jtp_mon15.png",  1+(2*MON_HUGE_DX),   1+(1*MON_HUGE_DY),  MON_HUGE_DX-4,MON_HUGE_DY-4},

/* cursors: they're actually normal tiles, so they're in here too */
/* statues: NAME                   FILENAME          top left x coord     top left y coord    width       height*/
{T_CURSOR, "CURSOR_NORMAL",        "jtp_mou5.png",   1,                   1,                  16,          23},
{T_CURSOR, "CURSOR_SCROLLRIGHT",   "jtp_mou5.png",   21,                  1,                  19,          22},
{T_CURSOR, "CURSOR_SCROLLLEFT",    "jtp_mou5.png",   44,                  1,                  19,          22},
{T_CURSOR, "CURSOR_SCROLLUP",      "jtp_mou5.png",   67,                  1,                  22,          19},
{T_CURSOR, "CURSOR_SCROLLDOWN",    "jtp_mou5.png",   93,                  1,                  22,          19},
{T_CURSOR, "CURSOR_TARGET_GREEN",  "jtp_mou5.png",   119,                 1,                  50,          25},
{T_CURSOR, "CURSOR_TARGET_RED",    "jtp_mou5.png",   173,                 1,                  50,          25},
{T_CURSOR, "CURSOR_TARGET_INVALID","jtp_mou5.png",   227,                 1,                  46,          25},
{T_CURSOR, "CURSOR_TARGET_HELP",   "jtp_mou5.png",   1,                   30,                 50,          49},
{T_CURSOR, "CURSOR_HOURGLASS",     "jtp_mou5.png",   277,                 1,                  29,          32},
{T_CURSOR, "CURSOR_OPENDOOR",      "jtp_mou5.png",   310,                 1,                  32,          29},
{T_CURSOR, "CURSOR_STAIRS",        "jtp_mou5.png",   346,                 1,                  37,          34},
{T_CURSOR, "CURSOR_GOBLET",        "jtp_mou5.png",   312,                 34,                 24,          34}
};

const int num_tiledefs = sizeof(tiledefinitions)/sizeof(struct tiledef);
const int num_objclassmaps = sizeof(objclassmap)/sizeof(struct classdefmap);
const int num_monclassmaps = sizeof(monclassmap)/sizeof(struct classdefmap);
const int num_eqmapentries = sizeof(tilemap)/sizeof(struct eqmap);
const int num_floor_edges = sizeof(floor_edges)/sizeof(char*);
const int num_fstyles = sizeof(fstyles)/sizeof(struct floorstyles);
const int num_wall_styles = sizeof(wall_styles)/sizeof(char*);

